# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import datetime
import os
import stat
import subprocess
import sys
import time
import threading
from contextlib import contextmanager

import pyomo.environ as pyo
from pyomo.common.envvar import is_windows
from pyomo.common.fileutils import ExecutableData
from pyomo.common.config import ConfigDict, ADVANCED_OPTION
from pyomo.common.errors import ApplicationError, MouseTrap
from pyomo.common.log import LoggingIntercept
from pyomo.common.tee import capture_output
from pyomo.common.timing import HierarchicalTimer
import pyomo.contrib.solver.solvers.ipopt as ipopt
from pyomo.contrib.solver.common.util import NoSolutionError, NoOptimalSolutionError
from pyomo.contrib.solver.common.results import TerminationCondition, SolutionStatus
from pyomo.contrib.solver.common.factory import SolverFactory
from pyomo.common import unittest, Executable
from pyomo.common.tempfiles import TempfileManager
from pyomo.repn.plugins.nl_writer import NLWriter, NLWriterInfo
from pyomo.repn.util import FileDeterminism

ipopt_available = ipopt.Ipopt().available()


@contextmanager
def windows_tee_buffer(size=1 << 20):
    """Temporarily increase TeeStream buffer size on Windows"""
    if not sys.platform.startswith("win"):
        # Only windows has an issue
        yield
        return
    import pyomo.common.tee as tee

    old = tee._pipe_buffersize
    tee._pipe_buffersize = size
    try:
        yield
    finally:
        tee._pipe_buffersize = old


@unittest.pytest.mark.solver("ipopt")
class TestIpoptSolverConfig(unittest.TestCase):
    def test_default_instantiation(self):
        config = ipopt.IpoptConfig()
        # Should be inherited
        self.assertIsNone(config._description)
        self.assertEqual(config._visibility, 0)
        self.assertFalse(config.tee)
        self.assertTrue(config.load_solutions)
        self.assertTrue(config.raise_exception_on_nonoptimal_result)
        self.assertFalse(config.symbolic_solver_labels)
        self.assertIsNone(config.timer)
        self.assertIsNone(config.threads)
        self.assertIsNone(config.time_limit)
        # Unique to this object
        self.assertIsInstance(config.executable, type(Executable('path')))
        self.assertIsInstance(config.writer_config, type(NLWriter.CONFIG()))

    def test_custom_instantiation(self):
        config = ipopt.IpoptConfig(description="A description")
        config.tee = True
        self.assertTrue(config.tee)
        self.assertEqual(config._description, "A description")
        self.assertIsNone(config.time_limit)
        # Default should be `ipopt`
        self.assertEqual('ipopt', config.executable._registered_name)
        if ipopt_available:
            self.assertIsNotNone(config.executable.path())
            self.assertIn('ipopt', str(config.executable))
        # Set to a totally bogus path
        config.executable = Executable('/bogus/path')
        self.assertIsNone(config.executable.executable)
        self.assertFalse(config.executable.available())


@unittest.pytest.mark.solver("ipopt")
class TestIpoptSolutionLoader(unittest.TestCase):
    def test_get_reduced_costs_error(self):
        loader = ipopt.IpoptSolutionLoader(
            ipopt.ASLSolFileData(), NLWriterInfo(eliminated_vars=[1])
        )
        with self.assertRaisesRegex(
            MouseTrap, "Complete reduced costs are not available"
        ):
            loader.get_reduced_costs()

    def test_get_duals_error(self):
        loader = ipopt.IpoptSolutionLoader(
            ipopt.ASLSolFileData(), NLWriterInfo(eliminated_vars=[1])
        )
        with self.assertRaisesRegex(MouseTrap, "Complete duals are not available"):
            loader.get_duals()


@unittest.pytest.mark.solver("ipopt")
class TestIpoptInterface(unittest.TestCase):
    @unittest.skipIf(not ipopt_available, "The 'ipopt' command is not available")
    def test_command_line_options(self):
        result = subprocess.run(
            [str(ipopt.Ipopt.CONFIG.executable), '-='],
            capture_output=True,
            text=True,
            check=True,
        )
        output = result.stdout
        options = []
        for line in output.splitlines():
            option_name = line.split()[0] if line.strip() else ''
            if option_name:
                options.append(option_name)
        self.assertEqual(sorted(ipopt.ipopt_command_line_options), sorted(options))

    def test_class_member_list(self):
        opt = ipopt.Ipopt()
        expected_list = [
            'CONFIG',
            'config',
            'api_version',
            'available',
            'has_linear_solver',
            'is_persistent',
            'solve',
            'version',
            'name',
        ]
        method_list = [method for method in dir(opt) if not method.startswith('_')]
        self.assertEqual(sorted(expected_list), sorted(method_list))

    def test_default_instantiation(self):
        opt = ipopt.Ipopt()
        self.assertFalse(opt.is_persistent())
        self.assertEqual(opt.name, 'ipopt')
        self.assertEqual(opt.CONFIG, opt.config)
        if ipopt_available:
            self.assertIsNotNone(opt.version())
            self.assertTrue(opt.available())
        else:
            self.assertIsNone(opt.version())
            self.assertFalse(opt.available())

    def test_context_manager(self):
        with ipopt.Ipopt() as opt:
            self.assertFalse(opt.is_persistent())
            self.assertEqual(opt.name, 'ipopt')
            self.assertEqual(opt.CONFIG, opt.config)
            if ipopt_available:
                self.assertIsNotNone(opt.version())
                self.assertTrue(opt.available())
            else:
                self.assertIsNone(opt.version())
                self.assertFalse(opt.available())

    def test_get_version(self):
        if ipopt_available:
            ver = ipopt.Ipopt().version()
            self.assertIsInstance(ver, tuple)
            self.assertEqual(len(ver), 3)
            self.assertTrue(all(isinstance(_, int) for _ in ver))

        _cache = ipopt.Ipopt._exe_cache
        try:
            with TempfileManager.new_context() as TMP:
                dname = TMP.mkdtemp()

                ipopt.Ipopt._exe_cache = {}
                fname = os.path.join(dname, 'test1')
                solver = ipopt.Ipopt(executable=fname)
                self.assertEqual({}, ipopt.Ipopt._exe_cache)
                self.assertEqual(ipopt.Availability.NotFound, solver.available())
                self.assertIsNone(solver.version())
                self.assertEqual({None: None}, ipopt.Ipopt._exe_cache)

                # the rest of this test is designed to work on *nix:
                if sys.platform.startswith("win"):
                    return

                ipopt.Ipopt._exe_cache = {}
                fname = os.path.join(dname, 'test2')
                with open(fname, 'w') as F:
                    F.write(f"#!{sys.executable}\nimport sys\nsys.exit(0)\n")
                solver = ipopt.Ipopt(executable=fname)
                self.assertEqual({}, ipopt.Ipopt._exe_cache)
                self.assertEqual(ipopt.Availability.NotFound, solver.available())
                self.assertIsNone(solver.version())
                self.assertEqual({None: None}, ipopt.Ipopt._exe_cache)

                # Found an executable, but --version errors
                ipopt.Ipopt._exe_cache = {}
                fname = os.path.join(dname, 'test3')
                with open(fname, 'w') as F:
                    F.write(f"#!{sys.executable}\nimport sys\nsys.exit(1)\n")
                os.chmod(fname, stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR)
                solver = ipopt.Ipopt(executable=fname)
                self.assertEqual({}, ipopt.Ipopt._exe_cache)
                self.assertEqual(ipopt.Availability.NotFound, solver.available())
                self.assertIsNone(solver.version())
                self.assertEqual({fname: None}, ipopt.Ipopt._exe_cache)

                # Found an executable, but --version doesn't return anything
                ipopt.Ipopt._exe_cache = {}
                fname = os.path.join(dname, 'test4')
                with open(fname, 'w') as F:
                    F.write(f"#!{sys.executable}\nimport sys\nsys.exit(0)\n")
                os.chmod(fname, stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR)
                solver = ipopt.Ipopt(executable=fname)
                self.assertEqual({}, ipopt.Ipopt._exe_cache)
                self.assertEqual(ipopt.Availability.NotFound, solver.available())
                self.assertIsNone(solver.version())
                self.assertEqual({fname: None}, ipopt.Ipopt._exe_cache)

                # Missing "ipopt"
                ipopt.Ipopt._exe_cache = {}
                fname = os.path.join(dname, 'test5')
                with open(fname, 'w') as F:
                    F.write(
                        f"#!{sys.executable}\nprint('cbc 1.2.3 ASL')\n"
                        "import sys\nsys.exit(0)\n"
                    )
                os.chmod(fname, stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR)
                solver = ipopt.Ipopt(executable=fname)
                self.assertEqual({}, ipopt.Ipopt._exe_cache)
                self.assertEqual(ipopt.Availability.NotFound, solver.available())
                self.assertIsNone(solver.version())
                self.assertEqual({fname: None}, ipopt.Ipopt._exe_cache)

                # The version doesn't parse correctly
                ipopt.Ipopt._exe_cache = {}
                fname = os.path.join(dname, 'test6')
                with open(fname, 'w') as F:
                    F.write(
                        f"#!{sys.executable}\nprint('Ipopt 1.2.3a ASL')\n"
                        "import sys\nsys.exit(0)\n"
                    )
                os.chmod(fname, stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR)
                solver = ipopt.Ipopt(executable=fname)
                self.assertEqual({}, ipopt.Ipopt._exe_cache)
                self.assertEqual(ipopt.Availability.NotFound, solver.available())
                self.assertIsNone(solver.version())
                self.assertEqual({fname: None}, ipopt.Ipopt._exe_cache)

                # This looks like an Ipopt solver...
                ipopt.Ipopt._exe_cache = {}
                fname = os.path.join(dname, 'test7')
                with open(fname, 'w') as F:
                    F.write(
                        f"#!{sys.executable}\nprint('Ipopt 1.2.3 ASL')\n"
                        "import sys\nsys.exit(0)\n"
                    )
                os.chmod(fname, stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR)
                solver = ipopt.Ipopt(executable=fname)
                self.assertEqual({}, ipopt.Ipopt._exe_cache)
                self.assertEqual(ipopt.Availability.FullLicense, solver.available())
                self.assertEqual(solver.version(), (1, 2, 3))
                self.assertEqual({fname: (1, 2, 3)}, ipopt.Ipopt._exe_cache)

        finally:
            ipopt.Ipopt._exe_cache = _cache

    def test_parse_output(self):
        # Old ipopt style (<=3.13)
        # Note: we are removing the URLs from the baseline because they
        # do not impact the test (and checking the URLs is fragile)
        output = """Ipopt 3.13.2:

******************************************************************************
This program contains Ipopt, a library for large-scale nonlinear optimization.
 Ipopt is released as open source code under the Eclipse Public License (EPL).
         For more information visit

This version of Ipopt was compiled from source code available at
    https://github.com/IDAES/Ipopt as part of the Institute for the Design of
    Advanced Energy Systems Process Systems Engineering Framework (IDAES PSE
    Framework) Copyright (c) 2018-2019. See https://github.com/IDAES/idaes-pse.

This version of Ipopt was compiled using HSL, a collection of Fortran codes
    for large-scale scientific computation.  All technical papers, sales and
    publicity material resulting from use of the HSL codes within IPOPT must
    contain the following acknowledgement:
        HSL, a collection of Fortran codes for large-scale scientific
        computation. See
******************************************************************************

This is Ipopt version 3.13.2, running with linear solver ma27.

Number of nonzeros in equality constraint Jacobian...:        0
Number of nonzeros in inequality constraint Jacobian.:        0
Number of nonzeros in Lagrangian Hessian.............:        3

Total number of variables............................:        2
                     variables with only lower bounds:        0
                variables with lower and upper bounds:        0
                     variables with only upper bounds:        0
Total number of equality constraints.................:        0
Total number of inequality constraints...............:        0
        inequality constraints with only lower bounds:        0
   inequality constraints with lower and upper bounds:        0
        inequality constraints with only upper bounds:        0

iter    objective    inf_pr   inf_du lg(mu)  ||d||  lg(rg) alpha_du alpha_pr  ls
   0  5.6500000e+01 0.00e+00 1.00e+02  -1.0 0.00e+00    -  0.00e+00 0.00e+00   0
   1  2.4669972e-01 0.00e+00 2.22e-01  -1.0 7.40e-01    -  1.00e+00 1.00e+00f  1
   2  1.6256267e-01 0.00e+00 2.04e+00  -1.7 1.48e+00    -  1.00e+00 2.50e-01f  3
   3  8.6119444e-02 0.00e+00 1.08e+00  -1.7 2.36e-01    -  1.00e+00 1.00e+00f  1
   4  4.3223836e-02 0.00e+00 1.23e+00  -1.7 2.61e-01    -  1.00e+00 1.00e+00f  1
   5  1.5610508e-02 0.00e+00 3.54e-01  -1.7 1.18e-01    -  1.00e+00 1.00e+00f  1
   6  5.3544798e-03 0.00e+00 5.51e-01  -1.7 1.67e-01    -  1.00e+00 1.00e+00f  1
   7  6.1281576e-04 0.00e+00 5.19e-02  -1.7 3.87e-02    -  1.00e+00 1.00e+00f  1
   8  2.8893076e-05 0.00e+00 4.52e-02  -2.5 4.53e-02    -  1.00e+00 1.00e+00f  1
   9  3.4591761e-08 0.00e+00 3.80e-04  -2.5 3.18e-03    -  1.00e+00 1.00e+00f  1
iter    objective    inf_pr   inf_du lg(mu)  ||d||  lg(rg) alpha_du alpha_pr  ls
  10  1.2680803e-13 0.00e+00 3.02e-06  -5.7 3.62e-04    -  1.00e+00 1.00e+00f  1
  11  7.0136460e-25 0.00e+00 1.72e-12  -8.6 2.13e-07    -  1.00e+00 1.00e+00f  1

Number of Iterations....: 11

                                   (scaled)                 (unscaled)
Objective...............:   1.5551321399859192e-25    7.0136459513364959e-25
Dual infeasibility......:   1.7239720368203862e-12    7.7751138860599418e-12
Constraint violation....:   0.0000000000000000e+00    0.0000000000000000e+00
Complementarity.........:   0.0000000000000000e+00    0.0000000000000000e+00
Overall NLP error.......:   1.7239720368203862e-12    7.7751138860599418e-12


Number of objective function evaluations             = 18
Number of objective gradient evaluations             = 12
Number of equality constraint evaluations            = 0
Number of inequality constraint evaluations          = 0
Number of equality constraint Jacobian evaluations   = 0
Number of inequality constraint Jacobian evaluations = 0
Number of Lagrangian Hessian evaluations             = 11
Total CPU secs in IPOPT (w/o function evaluations)   =      0.000
Total CPU secs in NLP function evaluations           =      0.000

EXIT: Optimal Solution Found.

        """
        parsed_output = ipopt.Ipopt()._parse_ipopt_output(output)
        self.assertEqual(
            {
                'iters': 11,
                'iteration_log': [
                    {
                        'iter': 0,
                        'objective': 56.5,
                        'inf_pr': 0.0,
                        'inf_du': 100.0,
                        'lg_mu': -1.0,
                        'd_norm': 0.0,
                        'lg_rg': None,
                        'alpha_du': 0.0,
                        'alpha_pr': 0.0,
                        'ls': 0,
                        'restoration': False,
                        'step_acceptance': None,
                    },
                    {
                        'iter': 1,
                        'objective': 0.24669972,
                        'inf_pr': 0.0,
                        'inf_du': 0.222,
                        'lg_mu': -1.0,
                        'd_norm': 0.74,
                        'lg_rg': None,
                        'alpha_du': 1.0,
                        'alpha_pr': 1.0,
                        'ls': 1,
                        'restoration': False,
                        'step_acceptance': 'f',
                    },
                    {
                        'iter': 2,
                        'objective': 0.16256267,
                        'inf_pr': 0.0,
                        'inf_du': 2.04,
                        'lg_mu': -1.7,
                        'd_norm': 1.48,
                        'lg_rg': None,
                        'alpha_du': 1.0,
                        'alpha_pr': 0.25,
                        'ls': 3,
                        'restoration': False,
                        'step_acceptance': 'f',
                    },
                    {
                        'iter': 3,
                        'objective': 0.086119444,
                        'inf_pr': 0.0,
                        'inf_du': 1.08,
                        'lg_mu': -1.7,
                        'd_norm': 0.236,
                        'lg_rg': None,
                        'alpha_du': 1.0,
                        'alpha_pr': 1.0,
                        'ls': 1,
                        'restoration': False,
                        'step_acceptance': 'f',
                    },
                    {
                        'iter': 4,
                        'objective': 0.043223836,
                        'inf_pr': 0.0,
                        'inf_du': 1.23,
                        'lg_mu': -1.7,
                        'd_norm': 0.261,
                        'lg_rg': None,
                        'alpha_du': 1.0,
                        'alpha_pr': 1.0,
                        'ls': 1,
                        'restoration': False,
                        'step_acceptance': 'f',
                    },
                    {
                        'iter': 5,
                        'objective': 0.015610508,
                        'inf_pr': 0.0,
                        'inf_du': 0.354,
                        'lg_mu': -1.7,
                        'd_norm': 0.118,
                        'lg_rg': None,
                        'alpha_du': 1.0,
                        'alpha_pr': 1.0,
                        'ls': 1,
                        'restoration': False,
                        'step_acceptance': 'f',
                    },
                    {
                        'iter': 6,
                        'objective': 0.0053544798,
                        'inf_pr': 0.0,
                        'inf_du': 0.551,
                        'lg_mu': -1.7,
                        'd_norm': 0.167,
                        'lg_rg': None,
                        'alpha_du': 1.0,
                        'alpha_pr': 1.0,
                        'ls': 1,
                        'restoration': False,
                        'step_acceptance': 'f',
                    },
                    {
                        'iter': 7,
                        'objective': 0.00061281576,
                        'inf_pr': 0.0,
                        'inf_du': 0.0519,
                        'lg_mu': -1.7,
                        'd_norm': 0.0387,
                        'lg_rg': None,
                        'alpha_du': 1.0,
                        'alpha_pr': 1.0,
                        'ls': 1,
                        'restoration': False,
                        'step_acceptance': 'f',
                    },
                    {
                        'iter': 8,
                        'objective': 2.8893076e-05,
                        'inf_pr': 0.0,
                        'inf_du': 0.0452,
                        'lg_mu': -2.5,
                        'd_norm': 0.0453,
                        'lg_rg': None,
                        'alpha_du': 1.0,
                        'alpha_pr': 1.0,
                        'ls': 1,
                        'restoration': False,
                        'step_acceptance': 'f',
                    },
                    {
                        'iter': 9,
                        'objective': 3.4591761e-08,
                        'inf_pr': 0.0,
                        'inf_du': 0.00038,
                        'lg_mu': -2.5,
                        'd_norm': 0.00318,
                        'lg_rg': None,
                        'alpha_du': 1.0,
                        'alpha_pr': 1.0,
                        'ls': 1,
                        'restoration': False,
                        'step_acceptance': 'f',
                    },
                    {
                        'iter': 10,
                        'objective': 1.2680803e-13,
                        'inf_pr': 0.0,
                        'inf_du': 3.02e-06,
                        'lg_mu': -5.7,
                        'd_norm': 0.000362,
                        'lg_rg': None,
                        'alpha_du': 1.0,
                        'alpha_pr': 1.0,
                        'ls': 1,
                        'restoration': False,
                        'step_acceptance': 'f',
                    },
                    {
                        'iter': 11,
                        'objective': 7.013646e-25,
                        'inf_pr': 0.0,
                        'inf_du': 1.72e-12,
                        'lg_mu': -8.6,
                        'd_norm': 2.13e-07,
                        'lg_rg': None,
                        'alpha_du': 1.0,
                        'alpha_pr': 1.0,
                        'ls': 1,
                        'restoration': False,
                        'step_acceptance': 'f',
                    },
                ],
                'incumbent_objective': 7.013645951336496e-25,
                'dual_infeasibility': 7.775113886059942e-12,
                'constraint_violation': 0.0,
                'complementarity_error': 0.0,
                'overall_nlp_error': 7.775113886059942e-12,
                'final_scaled_results': {
                    'incumbent_objective': 1.5551321399859192e-25,
                    'dual_infeasibility': 1.7239720368203862e-12,
                    'constraint_violation': 0.0,
                    'complementarity_error': 0.0,
                    'overall_nlp_error': 1.7239720368203862e-12,
                },
                'cpu_seconds': {
                    'IPOPT (w/o function evaluations)': 0.0,
                    'NLP function evaluations': 0.0,
                },
            },
            parsed_output,
        )

        # New ipopt style (3.14+)
        output = """******************************************************************************
This program contains Ipopt, a library for large-scale nonlinear optimization.
 Ipopt is released as open source code under the Eclipse Public License (EPL).
         For more information visit
******************************************************************************

This is Ipopt version 3.14.17, running with linear solver ma27.

Number of nonzeros in equality constraint Jacobian...:        0
Number of nonzeros in inequality constraint Jacobian.:        0
Number of nonzeros in Lagrangian Hessian.............:        3

Total number of variables............................:        2
                     variables with only lower bounds:        0
                variables with lower and upper bounds:        0
                     variables with only upper bounds:        0
Total number of equality constraints.................:        0
Total number of inequality constraints...............:        0
        inequality constraints with only lower bounds:        0
   inequality constraints with lower and upper bounds:        0
        inequality constraints with only upper bounds:        0

iter    objective    inf_pr   inf_du lg(mu)  ||d||  lg(rg) alpha_du alpha_pr  ls
   0  5.6500000e+01 0.00e+00 1.00e+02  -1.0 0.00e+00    -  0.00e+00 0.00e+00   0
   1  2.4669972e-01 0.00e+00 2.22e-01  -1.0 7.40e-01    -  1.00e+00 1.00e+00f  1
   2  1.6256267e-01 0.00e+00 2.04e+00  -1.7 1.48e+00    -  1.00e+00 2.50e-01f  3
   3  8.6119444e-02 0.00e+00 1.08e+00  -1.7 2.36e-01    -  1.00e+00 1.00e+00f  1
   4  4.3223836e-02 0.00e+00 1.23e+00  -1.7 2.61e-01    -  1.00e+00 1.00e+00f  1
   5  1.5610508e-02 0.00e+00 3.54e-01  -1.7 1.18e-01    -  1.00e+00 1.00e+00f  1
   6  5.3544798e-03 0.00e+00 5.51e-01  -1.7 1.67e-01    -  1.00e+00 1.00e+00f  1
   7  6.1281576e-04 0.00e+00 5.19e-02  -1.7 3.87e-02    -  1.00e+00 1.00e+00f  1
   8  2.8893076e-05 0.00e+00 4.52e-02  -2.5 4.53e-02    -  1.00e+00 1.00e+00f  1
   9  3.4591761e-08 0.00e+00 3.80e-04  -2.5 3.18e-03    -  1.00e+00 1.00e+00f  1
iter    objective    inf_pr   inf_du lg(mu)  ||d||  lg(rg) alpha_du alpha_pr  ls
  10  1.2680803e-13 0.00e+00 3.02e-06  -5.7 3.62e-04    -  1.00e+00 1.00e+00f  1
  11  7.0136460e-25 0.00e+00 1.72e-12  -8.6 2.13e-07    -  1.00e+00 1.00e+00f  1

Number of Iterations....: 11

                                   (scaled)                 (unscaled)
Objective...............:   1.5551321399859192e-25    7.0136459513364959e-25
Dual infeasibility......:   1.7239720368203862e-12    7.7751138860599418e-12
Constraint violation....:   0.0000000000000000e+00    0.0000000000000000e+00
Variable bound violation:   0.0000000000000000e+00    0.0000000000000000e+00
Complementarity.........:   0.0000000000000000e+00    0.0000000000000000e+00
Overall NLP error.......:   1.7239720368203862e-12    7.7751138860599418e-12


Number of objective function evaluations             = 18
Number of objective gradient evaluations             = 12
Number of equality constraint evaluations            = 0
Number of inequality constraint evaluations          = 0
Number of equality constraint Jacobian evaluations   = 0
Number of inequality constraint Jacobian evaluations = 0
Number of Lagrangian Hessian evaluations             = 11
Total seconds in IPOPT                               = 0.002

EXIT: Optimal Solution Found.

Ipopt 3.14.17: Optimal Solution Found
        """
        parsed_output = ipopt.Ipopt()._parse_ipopt_output(output)
        self.assertEqual(
            {
                'iters': 11,
                'iteration_log': [
                    {
                        'iter': 0,
                        'objective': 56.5,
                        'inf_pr': 0.0,
                        'inf_du': 100.0,
                        'lg_mu': -1.0,
                        'd_norm': 0.0,
                        'lg_rg': None,
                        'alpha_du': 0.0,
                        'alpha_pr': 0.0,
                        'ls': 0,
                        'restoration': False,
                        'step_acceptance': None,
                    },
                    {
                        'iter': 1,
                        'objective': 0.24669972,
                        'inf_pr': 0.0,
                        'inf_du': 0.222,
                        'lg_mu': -1.0,
                        'd_norm': 0.74,
                        'lg_rg': None,
                        'alpha_du': 1.0,
                        'alpha_pr': 1.0,
                        'ls': 1,
                        'restoration': False,
                        'step_acceptance': 'f',
                    },
                    {
                        'iter': 2,
                        'objective': 0.16256267,
                        'inf_pr': 0.0,
                        'inf_du': 2.04,
                        'lg_mu': -1.7,
                        'd_norm': 1.48,
                        'lg_rg': None,
                        'alpha_du': 1.0,
                        'alpha_pr': 0.25,
                        'ls': 3,
                        'restoration': False,
                        'step_acceptance': 'f',
                    },
                    {
                        'iter': 3,
                        'objective': 0.086119444,
                        'inf_pr': 0.0,
                        'inf_du': 1.08,
                        'lg_mu': -1.7,
                        'd_norm': 0.236,
                        'lg_rg': None,
                        'alpha_du': 1.0,
                        'alpha_pr': 1.0,
                        'ls': 1,
                        'restoration': False,
                        'step_acceptance': 'f',
                    },
                    {
                        'iter': 4,
                        'objective': 0.043223836,
                        'inf_pr': 0.0,
                        'inf_du': 1.23,
                        'lg_mu': -1.7,
                        'd_norm': 0.261,
                        'lg_rg': None,
                        'alpha_du': 1.0,
                        'alpha_pr': 1.0,
                        'ls': 1,
                        'restoration': False,
                        'step_acceptance': 'f',
                    },
                    {
                        'iter': 5,
                        'objective': 0.015610508,
                        'inf_pr': 0.0,
                        'inf_du': 0.354,
                        'lg_mu': -1.7,
                        'd_norm': 0.118,
                        'lg_rg': None,
                        'alpha_du': 1.0,
                        'alpha_pr': 1.0,
                        'ls': 1,
                        'restoration': False,
                        'step_acceptance': 'f',
                    },
                    {
                        'iter': 6,
                        'objective': 0.0053544798,
                        'inf_pr': 0.0,
                        'inf_du': 0.551,
                        'lg_mu': -1.7,
                        'd_norm': 0.167,
                        'lg_rg': None,
                        'alpha_du': 1.0,
                        'alpha_pr': 1.0,
                        'ls': 1,
                        'restoration': False,
                        'step_acceptance': 'f',
                    },
                    {
                        'iter': 7,
                        'objective': 0.00061281576,
                        'inf_pr': 0.0,
                        'inf_du': 0.0519,
                        'lg_mu': -1.7,
                        'd_norm': 0.0387,
                        'lg_rg': None,
                        'alpha_du': 1.0,
                        'alpha_pr': 1.0,
                        'ls': 1,
                        'restoration': False,
                        'step_acceptance': 'f',
                    },
                    {
                        'iter': 8,
                        'objective': 2.8893076e-05,
                        'inf_pr': 0.0,
                        'inf_du': 0.0452,
                        'lg_mu': -2.5,
                        'd_norm': 0.0453,
                        'lg_rg': None,
                        'alpha_du': 1.0,
                        'alpha_pr': 1.0,
                        'ls': 1,
                        'restoration': False,
                        'step_acceptance': 'f',
                    },
                    {
                        'iter': 9,
                        'objective': 3.4591761e-08,
                        'inf_pr': 0.0,
                        'inf_du': 0.00038,
                        'lg_mu': -2.5,
                        'd_norm': 0.00318,
                        'lg_rg': None,
                        'alpha_du': 1.0,
                        'alpha_pr': 1.0,
                        'ls': 1,
                        'restoration': False,
                        'step_acceptance': 'f',
                    },
                    {
                        'iter': 10,
                        'objective': 1.2680803e-13,
                        'inf_pr': 0.0,
                        'inf_du': 3.02e-06,
                        'lg_mu': -5.7,
                        'd_norm': 0.000362,
                        'lg_rg': None,
                        'alpha_du': 1.0,
                        'alpha_pr': 1.0,
                        'ls': 1,
                        'restoration': False,
                        'step_acceptance': 'f',
                    },
                    {
                        'iter': 11,
                        'objective': 7.013646e-25,
                        'inf_pr': 0.0,
                        'inf_du': 1.72e-12,
                        'lg_mu': -8.6,
                        'd_norm': 2.13e-07,
                        'lg_rg': None,
                        'alpha_du': 1.0,
                        'alpha_pr': 1.0,
                        'ls': 1,
                        'restoration': False,
                        'step_acceptance': 'f',
                    },
                ],
                'incumbent_objective': 7.013645951336496e-25,
                'dual_infeasibility': 7.775113886059942e-12,
                'constraint_violation': 0.0,
                'variable_bound_violation': 0.0,
                'complementarity_error': 0.0,
                'overall_nlp_error': 7.775113886059942e-12,
                'final_scaled_results': {
                    'incumbent_objective': 1.5551321399859192e-25,
                    'dual_infeasibility': 1.7239720368203862e-12,
                    'constraint_violation': 0.0,
                    'variable_bound_violation': 0.0,
                    'complementarity_error': 0.0,
                    'overall_nlp_error': 1.7239720368203862e-12,
                },
                'cpu_seconds': {'IPOPT': 0.002},
            },
            parsed_output,
        )

    def test_empty_output_parsing(self):
        with self.assertLogs(
            "pyomo.contrib.solver.solvers.ipopt", level="WARNING"
        ) as logs:
            ipopt.Ipopt()._parse_ipopt_output(output=None)
        self.assertIn(
            "Returned output from ipopt was empty. Cannot parse for additional data.",
            logs.output[0],
        )

    def test_parse_output_diagnostic_tags(self):
        output = """******************************************************************************
This program contains Ipopt, a library for large-scale nonlinear optimization.
 Ipopt is released as open source code under the Eclipse Public License (EPL).
         For more information visit [legacy Ipopt URL removed]

This version of Ipopt was compiled from source code available at
    https://github.com/IDAES/Ipopt as part of the Institute for the Design of
    Advanced Energy Systems Process Systems Engineering Framework (IDAES PSE
    Framework) Copyright (c) 2018-2019. See https://github.com/IDAES/idaes-pse.

This version of Ipopt was compiled using HSL, a collection of Fortran codes
    for large-scale scientific computation.  All technical papers, sales and
    publicity material resulting from use of the HSL codes within IPOPT must
    contain the following acknowledgement:
        HSL, a collection of Fortran codes for large-scale scientific
        computation. See http://www.hsl.rl.ac.uk/.
******************************************************************************

This is Ipopt version 3.13.2, running with linear solver ma57.

Number of nonzeros in equality constraint Jacobian...:    77541
Number of nonzeros in inequality constraint Jacobian.:        0
Number of nonzeros in Lagrangian Hessian.............:    51855

Total number of variables............................:    15468
                     variables with only lower bounds:     3491
                variables with lower and upper bounds:     5026
                     variables with only upper bounds:      186
Total number of equality constraints.................:    15417
Total number of inequality constraints...............:        0
        inequality constraints with only lower bounds:        0
   inequality constraints with lower and upper bounds:        0
        inequality constraints with only upper bounds:        0

iter    objective    inf_pr   inf_du lg(mu)  ||d||  lg(rg) alpha_du alpha_pr  ls
   0  4.3126674e+00 1.34e+00 1.00e+00  -5.0 0.00e+00    -  0.00e+00 0.00e+00   0
Reallocating memory for MA57: lfact (2247250)
   1r 4.3126674e+00 1.34e+00 9.99e+02   0.1 0.00e+00  -4.0 0.00e+00 3.29e-10R  2
   2r 3.0519246e+08 1.13e+00 9.90e+02   0.1 2.30e+02    -  2.60e-02 9.32e-03f  1
   3r 2.2712595e+09 1.69e+00 9.73e+02   0.1 2.23e+02    -  2.54e-02 1.71e-02f  1 Nhj
   4  2.2712065e+09 1.69e+00 1.37e+09  -5.0 3.08e+03    -  1.32e-05 1.17e-05f  1 q
   5  1.9062986e+09 1.55e+00 1.25e+09  -5.0 5.13e+03    -  1.19e-01 8.38e-02f  1
   6  1.7041594e+09 1.46e+00 1.18e+09  -5.0 5.66e+03    -  7.06e-02 5.45e-02f  1
   7  1.4763158e+09 1.36e+00 1.10e+09  -5.0 3.94e+03    -  2.30e-01 6.92e-02f  1
   8  8.5873108e+08 1.04e+00 8.41e+08  -5.0 2.38e+05    -  3.49e-06 2.37e-01f  1
   9  4.4215572e+08 7.45e-01 6.03e+08  -5.0 1.63e+06    -  7.97e-02 2.82e-01f  1
iter    objective    inf_pr   inf_du lg(mu)  ||d||  lg(rg) alpha_du alpha_pr  ls
  10  5.0251884e+01 1.65e-01 1.57e+04  -5.0 1.24e+06    -  3.92e-05 1.00e+00f  1
  11  4.9121733e+01 4.97e-02 4.68e+03  -5.0 8.11e+04    -  4.31e-02 7.01e-01h  1
  12  4.1483985e+01 2.24e-02 5.97e+03  -5.0 1.15e+06    -  5.93e-02 1.00e+00f  1
  13  3.5762585e+01 1.75e-02 5.00e+03  -5.0 1.03e+06    -  1.25e-01 1.00e+00f  1
  14  3.2291014e+01 1.08e-02 3.51e+03  -5.0 8.25e+05    -  6.68e-01 1.00e+00f  1
  15  3.2274630e+01 3.31e-05 1.17e+00  -5.0 4.26e+04    -  9.92e-01 1.00e+00h  1
  16  3.2274631e+01 7.45e-09 2.71e-03  -5.0 6.11e+02    -  8.97e-01 1.00e+00h  1
  17  3.2274635e+01 7.45e-09 2.35e-03  -5.0 2.71e+04    -  1.32e-01 1.00e+00f  1
  18  3.2274635e+01 7.45e-09 1.15e-04  -5.0 5.53e+03    -  9.51e-01 1.00e+00h  1
  19  3.2274635e+01 7.45e-09 2.84e-05  -5.0 4.41e+04    -  7.54e-01 1.00e+00f  1
iter    objective    inf_pr   inf_du lg(mu)  ||d||  lg(rg) alpha_du alpha_pr  ls
  20  3.2274635e+01 7.45e-09 8.54e-07  -5.0 1.83e+04    -  1.00e+00 1.00e+00h  1

Number of Iterations....: 20

                                   (scaled)                 (unscaled)
Objective...............:   3.2274635418964841e+01    3.2274635418964841e+01
Dual infeasibility......:   8.5365078678328669e-07    8.5365078678328669e-07
Constraint violation....:   8.0780625068607930e-13    7.4505805969238281e-09
Complementarity.........:   1.2275904566414160e-05    1.2275904566414160e-05
Overall NLP error.......:   1.2275904566414160e-05    1.2275904566414160e-05


Number of objective function evaluations             = 23
Number of objective gradient evaluations             = 20
Number of equality constraint evaluations            = 23
Number of inequality constraint evaluations          = 0
Number of equality constraint Jacobian evaluations   = 22
Number of inequality constraint Jacobian evaluations = 0
Number of Lagrangian Hessian evaluations             = 20
Total CPU secs in IPOPT (w/o function evaluations)   =     10.450
Total CPU secs in NLP function evaluations           =      1.651

EXIT: Optimal Solution Found.
    """
        parsed_output = ipopt.Ipopt()._parse_ipopt_output(output)
        self.assertEqual(
            {
                'iters': 20,
                'iteration_log': [
                    {
                        'iter': 0,
                        'objective': 4.3126674,
                        'inf_pr': 1.34,
                        'inf_du': 1.0,
                        'lg_mu': -5.0,
                        'd_norm': 0.0,
                        'lg_rg': None,
                        'alpha_du': 0.0,
                        'alpha_pr': 0.0,
                        'ls': 0,
                        'restoration': False,
                        'step_acceptance': None,
                    },
                    {
                        'iter': 1,
                        'objective': 4.3126674,
                        'inf_pr': 1.34,
                        'inf_du': 999.0,
                        'lg_mu': 0.1,
                        'd_norm': 0.0,
                        'lg_rg': -4.0,
                        'alpha_du': 0.0,
                        'alpha_pr': 3.29e-10,
                        'ls': 2,
                        'restoration': True,
                        'step_acceptance': 'R',
                    },
                    {
                        'iter': 2,
                        'objective': 305192460.0,
                        'inf_pr': 1.13,
                        'inf_du': 990.0,
                        'lg_mu': 0.1,
                        'd_norm': 230.0,
                        'lg_rg': None,
                        'alpha_du': 0.026,
                        'alpha_pr': 0.00932,
                        'ls': 1,
                        'restoration': True,
                        'step_acceptance': 'f',
                    },
                    {
                        'iter': 3,
                        'objective': 2271259500.0,
                        'inf_pr': 1.69,
                        'inf_du': 973.0,
                        'lg_mu': 0.1,
                        'd_norm': 223.0,
                        'lg_rg': None,
                        'alpha_du': 0.0254,
                        'alpha_pr': 0.0171,
                        'ls': 1,
                        'restoration': True,
                        'step_acceptance': 'f',
                        'diagnostic_tags': 'Nhj',
                    },
                    {
                        'iter': 4,
                        'objective': 2271206500.0,
                        'inf_pr': 1.69,
                        'inf_du': 1370000000.0,
                        'lg_mu': -5.0,
                        'd_norm': 3080.0,
                        'lg_rg': None,
                        'alpha_du': 1.32e-05,
                        'alpha_pr': 1.17e-05,
                        'ls': 1,
                        'restoration': False,
                        'step_acceptance': 'f',
                        'diagnostic_tags': 'q',
                    },
                    {
                        'iter': 5,
                        'objective': 1906298600.0,
                        'inf_pr': 1.55,
                        'inf_du': 1250000000.0,
                        'lg_mu': -5.0,
                        'd_norm': 5130.0,
                        'lg_rg': None,
                        'alpha_du': 0.119,
                        'alpha_pr': 0.0838,
                        'ls': 1,
                        'restoration': False,
                        'step_acceptance': 'f',
                    },
                    {
                        'iter': 6,
                        'objective': 1704159400.0,
                        'inf_pr': 1.46,
                        'inf_du': 1180000000.0,
                        'lg_mu': -5.0,
                        'd_norm': 5660.0,
                        'lg_rg': None,
                        'alpha_du': 0.0706,
                        'alpha_pr': 0.0545,
                        'ls': 1,
                        'restoration': False,
                        'step_acceptance': 'f',
                    },
                    {
                        'iter': 7,
                        'objective': 1476315800.0,
                        'inf_pr': 1.36,
                        'inf_du': 1100000000.0,
                        'lg_mu': -5.0,
                        'd_norm': 3940.0,
                        'lg_rg': None,
                        'alpha_du': 0.23,
                        'alpha_pr': 0.0692,
                        'ls': 1,
                        'restoration': False,
                        'step_acceptance': 'f',
                    },
                    {
                        'iter': 8,
                        'objective': 858731080.0,
                        'inf_pr': 1.04,
                        'inf_du': 841000000.0,
                        'lg_mu': -5.0,
                        'd_norm': 238000.0,
                        'lg_rg': None,
                        'alpha_du': 3.49e-06,
                        'alpha_pr': 0.237,
                        'ls': 1,
                        'restoration': False,
                        'step_acceptance': 'f',
                    },
                    {
                        'iter': 9,
                        'objective': 442155720.0,
                        'inf_pr': 0.745,
                        'inf_du': 603000000.0,
                        'lg_mu': -5.0,
                        'd_norm': 1630000.0,
                        'lg_rg': None,
                        'alpha_du': 0.0797,
                        'alpha_pr': 0.282,
                        'ls': 1,
                        'restoration': False,
                        'step_acceptance': 'f',
                    },
                    {
                        'iter': 10,
                        'objective': 50.251884,
                        'inf_pr': 0.165,
                        'inf_du': 15700.0,
                        'lg_mu': -5.0,
                        'd_norm': 1240000.0,
                        'lg_rg': None,
                        'alpha_du': 3.92e-05,
                        'alpha_pr': 1.0,
                        'ls': 1,
                        'restoration': False,
                        'step_acceptance': 'f',
                    },
                    {
                        'iter': 11,
                        'objective': 49.121733,
                        'inf_pr': 0.0497,
                        'inf_du': 4680.0,
                        'lg_mu': -5.0,
                        'd_norm': 81100.0,
                        'lg_rg': None,
                        'alpha_du': 0.0431,
                        'alpha_pr': 0.701,
                        'ls': 1,
                        'restoration': False,
                        'step_acceptance': 'h',
                    },
                    {
                        'iter': 12,
                        'objective': 41.483985,
                        'inf_pr': 0.0224,
                        'inf_du': 5970.0,
                        'lg_mu': -5.0,
                        'd_norm': 1150000.0,
                        'lg_rg': None,
                        'alpha_du': 0.0593,
                        'alpha_pr': 1.0,
                        'ls': 1,
                        'restoration': False,
                        'step_acceptance': 'f',
                    },
                    {
                        'iter': 13,
                        'objective': 35.762585,
                        'inf_pr': 0.0175,
                        'inf_du': 5000.0,
                        'lg_mu': -5.0,
                        'd_norm': 1030000.0,
                        'lg_rg': None,
                        'alpha_du': 0.125,
                        'alpha_pr': 1.0,
                        'ls': 1,
                        'restoration': False,
                        'step_acceptance': 'f',
                    },
                    {
                        'iter': 14,
                        'objective': 32.291014,
                        'inf_pr': 0.0108,
                        'inf_du': 3510.0,
                        'lg_mu': -5.0,
                        'd_norm': 825000.0,
                        'lg_rg': None,
                        'alpha_du': 0.668,
                        'alpha_pr': 1.0,
                        'ls': 1,
                        'restoration': False,
                        'step_acceptance': 'f',
                    },
                    {
                        'iter': 15,
                        'objective': 32.27463,
                        'inf_pr': 3.31e-05,
                        'inf_du': 1.17,
                        'lg_mu': -5.0,
                        'd_norm': 42600.0,
                        'lg_rg': None,
                        'alpha_du': 0.992,
                        'alpha_pr': 1.0,
                        'ls': 1,
                        'restoration': False,
                        'step_acceptance': 'h',
                    },
                    {
                        'iter': 16,
                        'objective': 32.274631,
                        'inf_pr': 7.45e-09,
                        'inf_du': 0.00271,
                        'lg_mu': -5.0,
                        'd_norm': 611.0,
                        'lg_rg': None,
                        'alpha_du': 0.897,
                        'alpha_pr': 1.0,
                        'ls': 1,
                        'restoration': False,
                        'step_acceptance': 'h',
                    },
                    {
                        'iter': 17,
                        'objective': 32.274635,
                        'inf_pr': 7.45e-09,
                        'inf_du': 0.00235,
                        'lg_mu': -5.0,
                        'd_norm': 27100.0,
                        'lg_rg': None,
                        'alpha_du': 0.132,
                        'alpha_pr': 1.0,
                        'ls': 1,
                        'restoration': False,
                        'step_acceptance': 'f',
                    },
                    {
                        'iter': 18,
                        'objective': 32.274635,
                        'inf_pr': 7.45e-09,
                        'inf_du': 0.000115,
                        'lg_mu': -5.0,
                        'd_norm': 5530.0,
                        'lg_rg': None,
                        'alpha_du': 0.951,
                        'alpha_pr': 1.0,
                        'ls': 1,
                        'restoration': False,
                        'step_acceptance': 'h',
                    },
                    {
                        'iter': 19,
                        'objective': 32.274635,
                        'inf_pr': 7.45e-09,
                        'inf_du': 2.84e-05,
                        'lg_mu': -5.0,
                        'd_norm': 44100.0,
                        'lg_rg': None,
                        'alpha_du': 0.754,
                        'alpha_pr': 1.0,
                        'ls': 1,
                        'restoration': False,
                        'step_acceptance': 'f',
                    },
                    {
                        'iter': 20,
                        'objective': 32.274635,
                        'inf_pr': 7.45e-09,
                        'inf_du': 8.54e-07,
                        'lg_mu': -5.0,
                        'd_norm': 18300.0,
                        'lg_rg': None,
                        'alpha_du': 1.0,
                        'alpha_pr': 1.0,
                        'ls': 1,
                        'restoration': False,
                        'step_acceptance': 'h',
                    },
                ],
                'incumbent_objective': 32.27463541896484,
                'dual_infeasibility': 8.536507867832867e-07,
                'constraint_violation': 7.450580596923828e-09,
                'complementarity_error': 1.227590456641416e-05,
                'overall_nlp_error': 1.227590456641416e-05,
                'final_scaled_results': {
                    'incumbent_objective': 32.27463541896484,
                    'dual_infeasibility': 8.536507867832867e-07,
                    'constraint_violation': 8.078062506860793e-13,
                    'complementarity_error': 1.227590456641416e-05,
                    'overall_nlp_error': 1.227590456641416e-05,
                },
                'cpu_seconds': {
                    'IPOPT (w/o function evaluations)': 10.45,
                    'NLP function evaluations': 1.651,
                },
            },
            parsed_output,
        )

    def test_parse_output_errors(self):
        output = """******************************************************************************
******************************************************************************

This is Ipopt version 3.13.2, running with linear solver ma57.

Number of nonzeros in equality constraint Jacobian...:    77541
Number of nonzeros in inequality constraint Jacobian.:        0
Number of nonzeros in Lagrangian Hessian.............:    51855

Total number of variables............................:    15468
                     variables with only lower bounds:     3491
                variables with lower and upper bounds:     5026
                     variables with only upper bounds:      186
Total number of equality constraints.................:    15417
Total number of inequality constraints...............:        0
        inequality constraints with only lower bounds:        0
   inequality constraints with lower and upper bounds:        0
        inequality constraints with only upper bounds:        0

iter    objective    inf_pr   inf_du lg(mu)  ||d||  lg(rg) alpha_du alpha_pr  ls
   0  4.3126674e+00 1.34e+00 1.00e+00  -5.0 0.00e+00    -  0.00e+00 0.00e+00   0
Reallocating memory for MA57: lfact (2247250)
   1r-4.3126674e+00 1.34e+00 9.99e+02   0.1 0.00e+00  -4.0 0.00e+00 3.29e-10R  2
  19t  3.2274635e+01 7.45e-09 2.84e-05  -5.0 4.41e+04    -  7.54e-01 1.00e+00f  1
iter    objective    inf_pr   inf_du lg(mu)  ||d||  lg(rg) alpha_du alpha_pr  ls
  20  3.2274635e+01f 7.45e-09 8.54e-07  -5.0 1.83e+04    -  1.00e+00 1.00e+00h  1

Number of Iterations....: 20

                                   (scaled)                 (unscaled)
Objective...............:   3.2274635418964841e+01    3.2274635418964841e+01
Dual infeasibility......:   8.5365078678328669e-07    8.5365078678328669e-07
Constraint violation....:   8.0780625068607930e-13    7.4505805969238281e-09
Complementarity.........:   1.2275904566414160e-05    1.2275904566414160e-05
Overall NLP error.......:   1.2275904566414160e-05    1.2275904566414160e-05


Number of objective function evaluations             = 23
Number of objective gradient evaluations             = 20
Number of equality constraint evaluations            = 23
Number of inequality constraint evaluations          = 0
Number of equality constraint Jacobian evaluations   = 22
Number of inequality constraint Jacobian evaluations = 0
Number of Lagrangian Hessian evaluations             = 20
Total CPU secs in IPOPT (w/o function evaluations)   =     10.450
Total CPU secs in NLP function evaluations           =      1.651

EXIT: Optimal Solution Found.
    """
        with LoggingIntercept() as LOG:
            parsed_output = ipopt.Ipopt()._parse_ipopt_output(output)
        self.assertEqual(
            """Error parsing Ipopt log entry:
\tinvalid literal for int() with base 10: '19t'
\t  19t  3.2274635e+01 7.45e-09 2.84e-05  -5.0 4.41e+04    -  7.54e-01 1.00e+00f  1
Error parsing Ipopt log entry:
\tcould not convert string to float: '3.2274635e+01f'
\t  20  3.2274635e+01f 7.45e-09 8.54e-07  -5.0 1.83e+04    -  1.00e+00 1.00e+00h  1
Total number of iteration records parsed 4 does not match the number of iterations (20) plus one.
""",
            LOG.getvalue(),
        )
        self.assertEqual(
            {
                'iters': 20,
                'iteration_log': [
                    {
                        'iter': 0,
                        'objective': 4.3126674,
                        'inf_pr': 1.34,
                        'inf_du': 1.0,
                        'lg_mu': -5.0,
                        'd_norm': 0.0,
                        'lg_rg': None,
                        'alpha_du': 0.0,
                        'alpha_pr': 0.0,
                        'ls': 0,
                        'restoration': False,
                        'step_acceptance': None,
                    },
                    {
                        'iter': 1,
                        'objective': -4.3126674,
                        'inf_pr': 1.34,
                        'inf_du': 999.0,
                        'lg_mu': 0.1,
                        'd_norm': 0.0,
                        'lg_rg': -4.0,
                        'alpha_du': 0.0,
                        'alpha_pr': 3.29e-10,
                        'ls': 2,
                        'restoration': True,
                        'step_acceptance': 'R',
                    },
                    {
                        'iter': '19t',
                        'objective': 32.274635,
                        'inf_pr': 7.45e-09,
                        'inf_du': 2.84e-05,
                        'lg_mu': -5.0,
                        'd_norm': 44100.0,
                        'lg_rg': None,
                        'alpha_du': 0.754,
                        'alpha_pr': 1.0,
                        'ls': 1,
                        'restoration': False,
                        'step_acceptance': 'f',
                    },
                    {
                        'iter': 20,
                        'objective': '3.2274635e+01f',
                        'inf_pr': 7.45e-09,
                        'inf_du': 8.54e-07,
                        'lg_mu': -5.0,
                        'd_norm': 18300.0,
                        'lg_rg': None,
                        'alpha_du': 1.0,
                        'alpha_pr': 1.0,
                        'ls': 1,
                        'restoration': False,
                        'step_acceptance': 'h',
                    },
                ],
                'incumbent_objective': 32.27463541896484,
                'dual_infeasibility': 8.536507867832867e-07,
                'constraint_violation': 7.450580596923828e-09,
                'complementarity_error': 1.227590456641416e-05,
                'overall_nlp_error': 1.227590456641416e-05,
                'final_scaled_results': {
                    'incumbent_objective': 32.27463541896484,
                    'dual_infeasibility': 8.536507867832867e-07,
                    'constraint_violation': 8.078062506860793e-13,
                    'complementarity_error': 1.227590456641416e-05,
                    'overall_nlp_error': 1.227590456641416e-05,
                },
                'cpu_seconds': {
                    'IPOPT (w/o function evaluations)': 10.45,
                    'NLP function evaluations': 1.651,
                },
            },
            parsed_output,
        )

    @unittest.skipIf(not ipopt_available, "The 'ipopt' command is not available")
    def test_has_linear_solver(self):
        opt = ipopt.Ipopt()
        self.assertTrue(
            any(
                map(
                    opt.has_linear_solver,
                    [
                        'mumps',
                        'ma27',
                        'ma57',
                        'ma77',
                        'ma86',
                        'ma97',
                        'pardiso',
                        'pardisomkl',
                        'spral',
                        'wsmp',
                    ],
                )
            )
        )
        self.assertFalse(opt.has_linear_solver('bogus_linear_solver'))

    @unittest.skipIf(sys.platform.startswith("win"), "Test requires *nix")
    def test_command_line(self):
        with TempfileManager.new_context() as tempfile:
            dname = tempfile.mkdtemp()
            exe = os.path.join(dname, 'mock')
            with open(exe, 'w') as F:
                F.write(f"""#!{sys.executable}
import sys
if sys.argv[1] == '--version':
    print('ipopt 1.2.3 ASL')
else:
    print('\\n'.join(sys.argv[1:]))
    sys.exit(1)
""")
            os.chmod(exe, stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR)
            opt = ipopt.Ipopt(executable=exe)

            m = pyo.ConcreteModel()
            m.x = pyo.Var()
            m.o = pyo.Objective(expr=m.x**2)

            opts = dict(
                raise_exception_on_nonoptimal_result=False, load_solutions=False
            )
            # No custom options, no file created. Plain and simple.
            with LoggingIntercept() as LOG:
                result = opt.solve(m, **opts)
            self.assertEqual("", LOG.getvalue())
            cmd = result.solver_log.splitlines()
            self.assertTrue(cmd[0].endswith('nl'))
            self.assertEqual(cmd[1:], ['-AMPL'])

            # Custom command line options
            with LoggingIntercept() as LOG:
                result = opt.solve(m, **opts, solver_options={'max_iter': 4})
            self.assertEqual("", LOG.getvalue())
            cmd = result.solver_log.splitlines()
            self.assertTrue(cmd[0].endswith('nl'))
            self.assertEqual(cmd[1:], ['-AMPL', 'max_iter=4'])

            # Custom command line options; threads generates a warning
            with LoggingIntercept() as LOG:
                result = opt.solve(
                    m, **opts, threads=10, solver_options={'max_iter': 4}
                )
            self.assertEqual(
                "The `threads=10` option was specified, "
                "but this is not used by Ipopt.\n",
                LOG.getvalue(),
            )
            cmd = result.solver_log.splitlines()
            self.assertTrue(cmd[0].endswith('nl'))
            self.assertEqual(cmd[1:], ['-AMPL', 'max_iter=4'])

            # Custom command line options; threads generates a warning... unless it's 1
            with LoggingIntercept() as LOG:
                result = opt.solve(m, **opts, threads=1, solver_options={'max_iter': 4})
            self.assertEqual("", LOG.getvalue())
            cmd = result.solver_log.splitlines()
            self.assertTrue(cmd[0].endswith('nl'))
            self.assertEqual(cmd[1:], ['-AMPL', 'max_iter=4'])

            # Let's see if we correctly parse config.time_limit
            with LoggingIntercept() as LOG:
                result = opt.solve(
                    m, **opts, time_limit=10, solver_options={'max_iter': 4}
                )
            self.assertEqual("", LOG.getvalue())
            cmd = result.solver_log.splitlines()
            self.assertTrue(cmd[0].endswith('nl'))
            self.assertEqual(cmd[1:], ['-AMPL', 'max_iter=4', 'max_cpu_time=10.0'])

            # Now let's do multiple command line options
            with LoggingIntercept() as LOG:
                result = opt.solve(
                    m, **opts, solver_options={'max_iter': 4, 'max_cpu_time': 20}
                )
            self.assertEqual("", LOG.getvalue())
            cmd = result.solver_log.splitlines()
            self.assertTrue(cmd[0].endswith('nl'))
            self.assertEqual(cmd[1:], ['-AMPL', 'max_cpu_time=20', 'max_iter=4'])

            # but top-level options override solver_options
            with LoggingIntercept() as LOG:
                result = opt.solve(
                    m,
                    **opts,
                    time_limit=10,
                    solver_options={'max_iter': 4, 'max_cpu_time': 20},
                )
            self.assertEqual("", LOG.getvalue())
            cmd = result.solver_log.splitlines()
            self.assertTrue(cmd[0].endswith('nl'))
            self.assertEqual(cmd[1:], ['-AMPL', 'max_cpu_time=10.0', 'max_iter=4'])

    def test_option_to_str(self):
        # int / float / str
        self.assertEqual('opt=5', ipopt._option_to_cmd('opt', 5))
        self.assertEqual('opt=5.0', ipopt._option_to_cmd('opt', 5.0))
        self.assertEqual('opt="5"', ipopt._option_to_cmd('opt', '5'))

        # If the string contains a quote, then the name needs to be
        # quoted
        self.assertEqual("opt=\"'model'\"", ipopt._option_to_cmd('opt', "'model'"))
        self.assertEqual("opt='\"model\"'", ipopt._option_to_cmd('opt', '"model"'))
        # but if it has both, we will error
        with self.assertRaisesRegex(ValueError, 'single and double'):
            ipopt._option_to_cmd('opt', '"\'model"')

    def test_process_options(self):
        solver = ipopt.Ipopt()
        with TempfileManager.new_context() as TMP:
            dname = TMP.mkdtemp()

            # test no options
            fname = os.path.join(dname, 'test1.txt')
            cmd = solver._process_options(fname, {})
            self.assertFalse(os.path.exists(fname))
            self.assertEqual([], cmd)

            # command-line only options
            fname = os.path.join(dname, 'test2.txt')
            cmd = solver._process_options(fname, {'bound_push': 'no', 'max_iter': 5})
            self.assertFalse(os.path.exists(fname))
            self.assertEqual(['bound_push="no"', 'max_iter=5'], cmd)

            # both command line and options file
            fname = os.path.join(dname, 'test3.txt')
            cmd = solver._process_options(
                fname, {'custom_option_2': 5, 'bound_push': 'no', 'custom_option_1': 3}
            )
            self.assertTrue(os.path.exists(fname))
            with open(fname, 'r') as F:
                self.assertEqual('custom_option_2 5\ncustom_option_1 3\n', F.read())
            if '"' in fname:
                fname = "'" + fname + "'"
            else:
                fname = '"' + fname + '"'
            self.assertEqual(['bound_push="no"', f'option_file_name={fname}'], cmd)

            # only options file
            fname = os.path.join(dname, 'test4.txt')
            cmd = solver._process_options(fname, {'custom_option_3': 3})
            self.assertTrue(os.path.exists(fname))
            with open(fname, 'r') as F:
                self.assertEqual('custom_option_3 3\n', F.read())
            if '"' in fname:
                fname = "'" + fname + "'"
            else:
                fname = '"' + fname + '"'
            self.assertEqual([f'option_file_name={fname}'], cmd)

            # illegal options
            fname = os.path.join(dname, 'test5.txt')
            with self.assertRaisesRegex(
                ValueError,
                "unallowed Ipopt option 'wantsol': "
                "The solver interface requires the sol file to be created",
            ):
                solver._process_options(fname, {'bogus': 3, 'wantsol': False})
            self.assertFalse(os.path.exists(fname))

            fname = os.path.join(dname, 'test5.txt')
            with self.assertRaisesRegex(
                ValueError,
                "unallowed Ipopt option 'option_file_name': "
                'Pyomo generates the ipopt options file as part of the `solve` '
                'method.  Add all options to config.solver_options instead',
            ):
                solver._process_options(
                    fname, {'bogus': 3, 'option_file_name': 'myfile.opt'}
                )
            self.assertFalse(os.path.exists(fname))

    def test_presolve_prove_infeasible(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var(bounds=(0, 5))
        m.c = pyo.Constraint(expr=m.x == 10)
        m.obj = pyo.Objective(expr=m.x)

        timer = HierarchicalTimer()
        solver = ipopt.Ipopt()
        results = solver.solve(
            m,
            timer=timer,
            load_solutions=False,
            raise_exception_on_nonoptimal_result=False,
        )
        self.assertEqual(results.solution_status, SolutionStatus.noSolution)
        self.assertEqual(
            results.termination_condition, TerminationCondition.provenInfeasible
        )
        cfg = results.solver_config
        del cfg.executable
        self.assertEqual(
            {
                'load_solutions': False,
                'raise_exception_on_nonoptimal_result': False,
                'solver_options': {},
                'symbolic_solver_labels': False,
                'tee': [],
                'threads': None,
                'time_limit': None,
                'timer': timer,
                'working_dir': None,
                'writer_config': {
                    'column_order': None,
                    'export_defined_variables': True,
                    'export_nonlinear_variables': None,
                    'file_determinism': FileDeterminism.ORDERED,
                    'linear_presolve': True,
                    'row_order': None,
                    'scale_model': True,
                    'show_section_timing': False,
                    'skip_trivial_constraints': True,
                    'symbolic_solver_labels': False,
                },
            },
            cfg.value(),
        )
        self.assertLess(results.timing_info.wall_time, 1)
        self.assertEqual(
            results.timing_info.start_timestamp.tzinfo, datetime.timezone.utc
        )
        self.assertLess(
            (
                datetime.datetime.now(datetime.timezone.utc)
                - results.timing_info.start_timestamp
            ).seconds,
            1,
        )
        del results.extra_info.base_file_name
        del results.solver_config
        del results.timing_info.wall_time
        del results.timing_info.start_timestamp
        self.assertEqual(
            {
                'extra_info': {'iteration_count': 0},
                'incumbent_objective': None,
                'objective_bound': None,
                'solution_loader': None,
                'solution_status': SolutionStatus.noSolution,
                'solver_log': None,
                'solver_name': 'ipopt',
                'solver_version': None,
                'termination_condition': TerminationCondition.provenInfeasible,
                'timing_info': {'timer': timer},
            },
            results.value(),
        )

        with self.assertRaisesRegex(
            NoSolutionError, "Solution loader does not currently have a valid solution."
        ):
            results = solver.solve(
                m, timer=timer, raise_exception_on_nonoptimal_result=False
            )
        with self.assertRaisesRegex(
            NoOptimalSolutionError, "Solver did not find the optimal solution."
        ):
            results = solver.solve(m, timer=timer)

    def test_presolve_solveModel(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var(bounds=(0, 50))
        m.c = pyo.Constraint(expr=m.x == 10)
        m.obj = pyo.Objective(expr=m.x)

        timer = HierarchicalTimer()
        solver = ipopt.Ipopt()
        results = solver.solve(m, timer=timer)
        self.assertEqual(results.solution_status, SolutionStatus.optimal)
        self.assertEqual(
            results.termination_condition,
            TerminationCondition.convergenceCriteriaSatisfied,
        )
        cfg = results.solver_config
        del results.solver_config
        del cfg.executable
        self.assertEqual(
            {
                'load_solutions': True,
                'raise_exception_on_nonoptimal_result': True,
                'solver_options': {},
                'symbolic_solver_labels': False,
                'tee': [],
                'threads': None,
                'time_limit': None,
                'timer': timer,
                'working_dir': None,
                'writer_config': {
                    'column_order': None,
                    'export_defined_variables': True,
                    'export_nonlinear_variables': None,
                    'file_determinism': FileDeterminism.ORDERED,
                    'linear_presolve': True,
                    'row_order': None,
                    'scale_model': True,
                    'show_section_timing': False,
                    'skip_trivial_constraints': True,
                    'symbolic_solver_labels': False,
                },
            },
            cfg.value(),
        )
        self.assertLess(results.timing_info.wall_time, 1)
        del results.timing_info.wall_time
        self.assertEqual(
            results.timing_info.start_timestamp.tzinfo, datetime.timezone.utc
        )
        self.assertLess(
            (
                datetime.datetime.now(datetime.timezone.utc)
                - results.timing_info.start_timestamp
            ).seconds,
            1,
        )
        del results.timing_info.start_timestamp
        del results.extra_info.base_file_name
        self.assertIsNotNone(results.solution_loader)
        del results.solution_loader
        self.assertEqual(
            {
                'extra_info': {'iteration_count': 0},
                'incumbent_objective': 10.0,
                'objective_bound': None,
                'solution_status': SolutionStatus.optimal,
                'solver_log': None,
                'solver_name': 'ipopt',
                'solver_version': None,
                'termination_condition': TerminationCondition.convergenceCriteriaSatisfied,
                'timing_info': {'timer': timer},
            },
            results.value(),
        )
        self.assertEqual(m.x.value, 10)

    def test_presolve_empty(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var(bounds=(0, 5))
        m.obj = pyo.Objective(expr=1)

        timer = HierarchicalTimer()
        solver = ipopt.Ipopt()
        results = solver.solve(
            m,
            timer=timer,
            load_solutions=False,
            raise_exception_on_nonoptimal_result=False,
        )
        self.assertEqual(results.solution_status, SolutionStatus.noSolution)
        self.assertEqual(results.termination_condition, TerminationCondition.emptyModel)
        cfg = results.solver_config
        del cfg.executable
        self.assertEqual(
            {
                'load_solutions': False,
                'raise_exception_on_nonoptimal_result': False,
                'solver_options': {},
                'symbolic_solver_labels': False,
                'tee': [],
                'threads': None,
                'time_limit': None,
                'timer': timer,
                'working_dir': None,
                'writer_config': {
                    'column_order': None,
                    'export_defined_variables': True,
                    'export_nonlinear_variables': None,
                    'file_determinism': FileDeterminism.ORDERED,
                    'linear_presolve': True,
                    'row_order': None,
                    'scale_model': True,
                    'show_section_timing': False,
                    'skip_trivial_constraints': True,
                    'symbolic_solver_labels': False,
                },
            },
            cfg.value(),
        )
        self.assertLess(results.timing_info.wall_time, 1)
        self.assertEqual(
            results.timing_info.start_timestamp.tzinfo, datetime.timezone.utc
        )
        self.assertLess(
            (
                datetime.datetime.now(datetime.timezone.utc)
                - results.timing_info.start_timestamp
            ).seconds,
            1,
        )
        del results.extra_info.base_file_name
        del results.solver_config
        del results.timing_info.wall_time
        del results.timing_info.start_timestamp
        self.assertEqual(
            {
                'extra_info': {'iteration_count': 0},
                'incumbent_objective': None,
                'objective_bound': None,
                'solution_loader': None,
                'solution_status': SolutionStatus.noSolution,
                'solver_log': None,
                'solver_name': 'ipopt',
                'solver_version': None,
                'termination_condition': TerminationCondition.emptyModel,
                'timing_info': {'timer': timer},
            },
            results.value(),
        )

        with self.assertRaisesRegex(
            NoSolutionError, "Solution loader does not currently have a valid solution."
        ):
            results = solver.solve(
                m, timer=timer, raise_exception_on_nonoptimal_result=False
            )
        with self.assertRaisesRegex(
            NoOptimalSolutionError, "Solver did not find the optimal solution."
        ):
            results = solver.solve(m, timer=timer)

    def test_file_collision(self):
        class mock_tempfile:
            def __init__(self):
                self.fd = None

            def new_context(self):
                return self

            def __enter__(self):
                return self

            def __exit__(self, et, ev, tb):
                if self.fd is not None:
                    os.close(self.fd)

            def mkstemp(self, suffix, prefix, dir, text, delete):
                fname = os.path.join(dir, "testfile" + suffix)
                self.fd = os.open(fname, os.O_CREAT | os.O_RDWR)
                return self.fd, fname

        m = pyo.ConcreteModel()
        orig_TempfileManager = ipopt.TempfileManager
        try:
            ipopt.TempfileManager = mock_tempfile()
            with TempfileManager.new_context() as tempfile:
                solver = ipopt.Ipopt()
                dname = tempfile.mkdtemp()
                for ext in ('.row', '.col', '.sol', '.opt'):
                    fname = os.path.join(dname, 'testfile' + ext)
                    open(fname, 'w').close()
                    with self.assertRaisesRegex(
                        RuntimeError,
                        f"Solver interface file "
                        + fname.replace('\\', '\\\\')
                        + " already exists",
                    ):
                        solver.solve(m, working_dir=dname)
                    os.unlink(fname)
        finally:
            ipopt.TempfileManager = orig_TempfileManager

    def test_bad_executable(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.o = pyo.Objective(expr=m.x**2)

        solver = ipopt.Ipopt()

        with TempfileManager.new_context() as tempfile:
            dname = tempfile.mkdtemp()
            exe = os.path.join(dname, 'ipopt')
            solver.config.executable = exe
            with self.assertRaisesRegex(ApplicationError, 'ipopt executable not found'):
                solver.solve(m)

            # The following is designed to run on *NIX
            if sys.platform.startswith("win"):
                return

            _cache = ipopt.Ipopt._exe_cache
            ipopt.Ipopt._exe_cache = {exe: (1, 2, 3)}
            try:
                with open(exe, 'w') as F:
                    F.write(f"#!{dname}/bad_interpreter\nsys.exit(1)\n")
                os.chmod(exe, stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR)
                solver.config.executable.rehash()
                with self.assertRaisesRegex(
                    ApplicationError,
                    f"Could not execute the command: \\['{exe}'.*"
                    f"Error message: .*No such file or directory: '{exe}'",
                ):
                    solver.solve(m)
            finally:
                ipopt.Ipopt._exe_cache = _cache


@unittest.skipIf(not ipopt_available, "The 'ipopt' command is not available")
@unittest.pytest.mark.solver("ipopt")
class TestIpopt(unittest.TestCase):
    def create_model(self):
        model = pyo.ConcreteModel()
        model.x = pyo.Var(initialize=1.5)
        model.y = pyo.Var(initialize=1.5)

        def rosenbrock(m):
            return (1.0 - m.x) ** 2 + 100.0 * (m.y - m.x**2) ** 2

        model.obj = pyo.Objective(rule=rosenbrock, sense=pyo.minimize)
        return model

    def test_ipopt_config(self):
        # Test default initialization
        config = ipopt.IpoptConfig()
        self.assertTrue(config.load_solutions)
        self.assertIsInstance(config.solver_options, ConfigDict)
        self.assertIsInstance(config.executable, ExecutableData)

        # Test custom initialization
        solver = SolverFactory('ipopt', executable='/path/to/exe')
        self.assertFalse(solver.config.tee)
        self.assertIsNone(solver.config.executable.path())
        self.assertTrue(solver.config.executable._registered_name.startswith('/path'))

    def test_ipopt_solve(self):
        # Gut check - does it solve?
        model = self.create_model()
        ipopt.Ipopt().solve(model)
        self.assertAlmostEqual(model.x.value, 1)
        self.assertAlmostEqual(model.y.value, 1)

    def test_ipopt_quiet_print_level(self):
        model = self.create_model()
        result = ipopt.Ipopt().solve(model, solver_options={'print_level': 0})
        # IPOPT doesn't tell us anything about the iters if the print level
        # is set to 0
        self.assertEqual(result.extra_info.iteration_count, None)
        self.assertFalse(hasattr(result.extra_info, 'iteration_log'))
        model = self.create_model()
        result = ipopt.Ipopt().solve(model, solver_options={'print_level': 3})
        # At a slightly higher level, we get some of the info, like
        # iteration count, but NOT iteration_log
        self.assertEqual(result.extra_info.iteration_count, 11)
        self.assertFalse(hasattr(result.extra_info, 'iteration_log'))

    def test_ipopt_loud_print_level(self):
        with windows_tee_buffer(1 << 20):
            model = self.create_model()
            result = ipopt.Ipopt().solve(model, solver_options={'print_level': 8})
            # Nothing unexpected should be in the results object at this point,
            # except that the solver_log is significantly longer
            self.assertEqual(result.extra_info.iteration_count, 11)
            self.assertEqual(result.incumbent_objective, 7.013645951336496e-25)
            self.assertIn('Optimal Solution Found', result.extra_info.solver_message)
            self.assertTrue(hasattr(result.extra_info, 'iteration_log'))
            model = self.create_model()
            result = ipopt.Ipopt().solve(model, solver_options={'print_level': 12})
            self.assertEqual(result.extra_info.iteration_count, 11)
            self.assertEqual(result.incumbent_objective, 7.013645951336496e-25)
            self.assertIn('Optimal Solution Found', result.extra_info.solver_message)
            self.assertTrue(hasattr(result.extra_info, 'iteration_log'))

    def test_ipopt_results(self):
        model = self.create_model()
        results = ipopt.Ipopt().solve(model)
        self.assertEqual(results.solver_name, 'ipopt')
        self.assertEqual(results.extra_info.iteration_count, 11)
        self.assertEqual(results.incumbent_objective, 7.013645951336496e-25)
        self.assertIn('Optimal Solution Found', results.extra_info.solver_message)

    def test_ipopt_results_display(self):
        model = self.create_model()
        results = ipopt.Ipopt().solve(model)
        # Do not show extra loud stuff
        with capture_output() as OUT:
            results.display()
        contents = OUT.getvalue()
        self.assertIn('termination_condition', contents)
        self.assertIn('solution_status', contents)
        self.assertIn('incumbent_objective', contents)
        self.assertNotIn('iteration_log', contents)
        # Now we want to see the iteration log
        with capture_output() as OUT:
            results.display(visibility=ADVANCED_OPTION)
        contents = OUT.getvalue()
        self.assertIn('termination_condition', contents)
        self.assertIn('solution_status', contents)
        self.assertIn('incumbent_objective', contents)
        self.assertIn('iteration_log', contents)

    def test_ipopt_timer_object(self):
        model = self.create_model()
        ipopt_instance = ipopt.Ipopt()
        results = ipopt_instance.solve(model)
        timing_info = results.timing_info
        if ipopt_instance.version()[0:1] <= (3, 13):
            # We are running an older version of IPOPT (<= 3.13)
            self.assertIn('IPOPT (w/o function evaluations)', timing_info.keys())
            self.assertIn('NLP function evaluations', timing_info.keys())
        else:
            # Newer version of IPOPT
            self.assertIn('IPOPT', timing_info.keys())

    def test_run_ipopt_options_file(self):
        # Check that the options file is getting to Ipopt: if we give it
        # an invalid option in the options file, ipopt will fail.  This
        # is important, as ipopt will NOT fail if you pass if an
        # option_file_name that does not exist.
        model = self.create_model()
        results = ipopt.Ipopt().solve(
            model,
            solver_options={'bogus_option': 5},
            raise_exception_on_nonoptimal_result=False,
            load_solutions=False,
        )
        self.assertEqual(results.termination_condition, TerminationCondition.error)
        self.assertEqual(results.solution_status, SolutionStatus.noSolution)
        self.assertIn('OPTION_INVALID', results.solver_log)

        # Verify the full command line once
        cmd = results.extra_info.command_line
        fname = results.extra_info.base_file_name
        self.assertEqual(
            [
                str(ipopt.Ipopt().config.executable),
                f'{fname}.nl',
                "-AMPL",
                f'option_file_name="{fname}.opt"',
            ],
            cmd,
        )

    def test_ipopt_working_dir(self):
        m = self.create_model()
        with TempfileManager.new_context() as tempfile:
            dname = tempfile.mkdtemp()
            working_dir = os.path.join(dname, 'testing')
            self.assertFalse(os.path.exists(working_dir))

            results = ipopt.Ipopt().solve(m, working_dir=working_dir)

            self.assertTrue(os.path.exists(working_dir))
            self.assertTrue(results.extra_info.base_file_name.startswith(working_dir))
            self.assertTrue(os.path.exists(results.extra_info.base_file_name + '.nl'))

    def test_load_solution_suffixes(self):
        m = self.create_model()
        m.x.lb = 0.6
        m.dual = pyo.Suffix(direction=pyo.Suffix.IMPORT)
        m.rc = pyo.Suffix(direction=pyo.Suffix.IMPORT)
        m.c = pyo.Constraint(expr=m.x == 2 * m.y)

        solver = ipopt.Ipopt()
        results = solver.solve(m, writer_config={'linear_presolve': False})
        o1 = results.extra_info.incumbent_objective

        self.assertAlmostEqual(m.x.value, 0.6, delta=1e-5)
        self.assertAlmostEqual(m.y.value, 0.3, delta=1e-5)
        self.assertEqual(len(m.dual), 1)
        self.assertAlmostEqual(m.dual[m.c], 6, delta=1e-5)
        self.assertEqual(len(m.rc), 2)
        self.assertAlmostEqual(m.rc[m.x], 7.6, delta=1e-5)
        self.assertEqual(m.rc[m.y], 0)

        m.scaling_factor = pyo.Suffix(direction=pyo.Suffix.EXPORT)
        m.scaling_factor[m.obj] = 10
        m.scaling_factor[m.c] = 5
        m.scaling_factor[m.x] = 7
        m.scaling_factor[m.y] = 3
        results = solver.solve(m, writer_config={'linear_presolve': False})

        o2 = results.extra_info.incumbent_objective
        self.assertAlmostEqual(o1, o2 / 10, delta=1e-5)

        self.assertAlmostEqual(m.x.value, 0.6, delta=1e-5)
        self.assertAlmostEqual(m.y.value, 0.3, delta=1e-5)
        self.assertEqual(len(m.dual), 1)
        self.assertAlmostEqual(m.dual[m.c], 6, delta=1e-5)
        self.assertEqual(len(m.rc), 2)
        self.assertAlmostEqual(m.rc[m.x], 7.6, delta=1e-5)
        self.assertEqual(m.rc[m.y], 0)

        m.x.lb = None
        m.y.ub = 0.25
        results = solver.solve(m, writer_config={'linear_presolve': False})

        self.assertAlmostEqual(m.x.value, 0.5, delta=1e-5)
        self.assertAlmostEqual(m.y.value, 0.25, delta=1e-5)
        self.assertEqual(len(m.dual), 1)
        self.assertAlmostEqual(m.dual[m.c], -1, delta=1e-5)
        self.assertEqual(len(m.rc), 2)
        self.assertEqual(m.rc[m.x], 0)
        self.assertAlmostEqual(m.rc[m.y], -2, delta=1e-5)

    def test_load_suffixes_infeasible_model(self):
        # This tests Issue #3807: Ipopt was failing to load duals /
        # reduced coses when the solver exited in restoration
        m = pyo.ConcreteModel()
        m.x = pyo.Var(initialize=2, bounds=(1, 10))
        m.c = pyo.Constraint(expr=(m.x == 0))

        solver = ipopt.Ipopt()
        solver.config.writer_config.linear_presolve = False
        solver.config.raise_exception_on_nonoptimal_result = False

        results = solver.solve(m)
        self.assertEqual(results.solution_status, SolutionStatus.infeasible)
        self.assertEqual(results.extra_info.iteration_log[-1]['restoration'], True)
        self.assertAlmostEqual(results.solution_loader.get_reduced_costs()[m.x], 1000)
        self.assertAlmostEqual(results.solution_loader.get_duals()[m.c], -1000)


@unittest.skipIf(not ipopt_available, "The 'ipopt' command is not available")
@unittest.pytest.mark.solver("ipopt")
class TestLegacyIpopt(unittest.TestCase):
    def create_model(self):
        model = pyo.ConcreteModel()
        model.x = pyo.Var(initialize=1.5)
        model.y = pyo.Var(initialize=1.5)

        @model.Objective(sense=pyo.minimize)
        def rosenbrock(m):
            return (1.0 - m.x) ** 2 + 100.0 * (m.y - m.x**2) ** 2

        return model

    def test_map_OF_options(self):
        model = self.create_model()

        with capture_output() as LOG:
            results = ipopt.LegacyIpoptSolver().solve(
                model,
                tee=True,
                solver_options={'OF_bogus_option': 5},
                load_solutions=False,
            )
        self.assertIn('OPTION_INVALID', LOG.getvalue())
        # Note: OF_ is stripped
        self.assertIn(
            'Read Option: "bogus_option". It is not a valid option', LOG.getvalue()
        )

        with self.assertRaisesRegex(ValueError, "unallowed Ipopt option 'wantsol'"):
            results = ipopt.LegacyIpoptSolver().solve(
                model,
                tee=True,
                solver_options={'OF_wantsol': False},
                load_solutions=False,
            )



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import code
import sys
import subprocess

from pyomo.common._command import pyomo_command
from pyomo.common.deprecation import deprecated
import pyomo.scripting.pyomo_parser


@pyomo_command('pyomo_python', "Launch script using Pyomo's python installation")
@deprecated(
    msg="The 'pyomo_python' command has been deprecated and will be removed",
    version="6.0",
)
def pyomo_python(args=None):
    if args is None:
        args = sys.argv[1:]
    if args is None or len(args) == 0:
        console = code.InteractiveConsole()
        console.interact('Pyomo Python Console\n' + sys.version)
    else:
        cmd = sys.executable + ' ' + ' '.join(args)
        subprocess.run(cmd)


@pyomo_command('pyomo', "The main command interface for Pyomo")
def pyomo(args=None):
    parser = pyomo.scripting.pyomo_parser.get_parser()
    if args is None:
        ret = parser.parse_args()
    else:
        ret = parser.parse_args(args)
    ret.func(ret)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""Implementation of MINLP problem in Assignment 6 of the Advanced PSE lecture at CMU.

Author: David Bernal <https://github.com/bernalde>

The expected optimal solution is 3.5.

Ref:
    IGNACIO GROSSMANN.
    CARNEGIE-MELLON UNIVERSITY , PITTSBURGH , PA.

    Problem type:    convex MINLP
            size:    3  binary variables
                     3  continuous variables
                     7  constraints

"""

from pyomo.environ import (
    Binary,
    ConcreteModel,
    Constraint,
    NonNegativeReals,
    Objective,
    RangeSet,
    Var,
    minimize,
    Block,
)
from pyomo.common.collections import ComponentMap
from pyomo.contrib.mindtpy.tests.MINLP_simple_grey_box import (
    GreyBoxModel,
    build_model_external,
)


class SimpleMINLP(ConcreteModel):
    """Convex MINLP problem Assignment 6 APSE."""

    def __init__(self, grey_box=False, *args, **kwargs):
        """Create the problem."""
        kwargs.setdefault('name', 'SimpleMINLP')
        if grey_box and GreyBoxModel is None:
            m = None
            return

        super(SimpleMINLP, self).__init__(*args, **kwargs)
        m = self

        """Set declarations"""
        I = m.I = RangeSet(1, 2, doc='continuous variables')
        J = m.J = RangeSet(1, 3, doc='discrete variables')

        # initial point information for discrete variables
        initY = {
            'sub1': {1: 1, 2: 1, 3: 1},
            'sub2': {1: 0, 2: 1, 3: 1},
            'sub3': {1: 1, 2: 0, 3: 1},
            'sub4': {1: 1, 2: 1, 3: 0},
            'sub5': {1: 0, 2: 0, 3: 0},
        }
        # initial point information for continuous variables
        initX = {1: 0, 2: 0}

        """Variable declarations"""
        # DISCRETE VARIABLES
        Y = m.Y = Var(J, domain=Binary, initialize=initY['sub2'])
        # CONTINUOUS VARIABLES
        X = m.X = Var(I, domain=NonNegativeReals, initialize=initX)

        """Constraint definitions"""
        # CONSTRAINTS
        m.const1 = Constraint(expr=(m.X[1] - 2) ** 2 - m.X[2] <= 0)
        m.const2 = Constraint(expr=m.X[1] - 2 * m.Y[1] >= 0)
        m.const3 = Constraint(expr=m.X[1] - m.X[2] - 4 * (1 - m.Y[2]) <= 0)
        m.const4 = Constraint(expr=m.X[1] - (1 - m.Y[1]) >= 0)
        m.const5 = Constraint(expr=m.X[2] - m.Y[2] >= 0)
        m.const6 = Constraint(expr=m.X[1] + m.X[2] >= 3 * m.Y[3])
        m.const7 = Constraint(expr=m.Y[1] + m.Y[2] + m.Y[3] >= 1)

        """Cost (objective) function definition"""
        m.objective = Objective(
            expr=Y[1] + 1.5 * Y[2] + 0.5 * Y[3] + X[1] ** 2 + X[2] ** 2, sense=minimize
        )

        if not grey_box:
            m.objective = Objective(
                expr=Y[1] + 1.5 * Y[2] + 0.5 * Y[3] + X[1] ** 2 + X[2] ** 2,
                sense=minimize,
            )
        else:

            def _model_i(b):
                build_model_external(b)

            m.my_block = Block(rule=_model_i)

            for i in m.I:

                def eq_inputX(m):
                    return m.X[i] == m.my_block.egb.inputs["X" + str(i)]

                con_name = "con_X_" + str(i)
                m.add_component(con_name, Constraint(expr=eq_inputX))

            for j in m.J:

                def eq_inputY(m):
                    return m.Y[j] == m.my_block.egb.inputs["Y" + str(j)]

                con_name = "con_Y_" + str(j)
                m.add_component(con_name, Constraint(expr=eq_inputY))

            # add objective
            m.objective = Objective(expr=m.my_block.egb.outputs['z'], sense=minimize)

        """Bound definitions"""
        # x (continuous) upper bounds
        x_ubs = {1: 4, 2: 4}
        for i, x_ub in x_ubs.items():
            X[i].setub(x_ub)

        m.optimal_value = 3.5
        m.optimal_solution = ComponentMap()
        m.optimal_solution[m.X[1]] = 1.0
        m.optimal_solution[m.X[2]] = 1.0
        m.optimal_solution[m.Y[1]] = 0.0
        m.optimal_solution[m.Y[2]] = 1.0
        m.optimal_solution[m.Y[3]] = 0.0



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Unit Tests for pyomo.opt.base.OS
#

import os

import pyomo.common.unittest as unittest
from pyomo.common.errors import ApplicationError
from pyomo.common.fileutils import this_file_dir

from pyomo.opt.base import UnknownSolver
from pyomo.opt.base.solvers import SolverFactory
from pyomo.opt.solver import SystemCallSolver

thisdir = os.path.normpath(this_file_dir())
exedirname = "exe_dir"
exedir = os.path.join(thisdir, exedirname)
user_home = os.path.normpath(os.path.expanduser("~"))
user_home_pos = exedir.find(user_home)
exedir_user = exedir
if user_home_pos >= 0:
    _test_user_exedir = os.path.join(
        user_home, exedir[user_home_pos + len(user_home) + 1 :]
    )
    if os.path.samefile(exedir, _test_user_exedir):
        exedir_user = os.path.join("~", exedir[user_home_pos + len(user_home) + 1 :])

notexe_nopath = "file_not_executable"
notexe_abspath = os.path.join(exedir, notexe_nopath)
notexe_abspath_user = os.path.join(exedir_user, notexe_nopath)
notexe_relpath = os.path.curdir + os.path.sep + exedirname + os.path.sep + notexe_nopath

isexe_nopath = "file_is_executable"
isexe_abspath = os.path.join(exedir, isexe_nopath)
isexe_abspath_user = os.path.join(exedir_user, isexe_nopath)
isexe_relpath = os.path.curdir + os.path.sep + exedirname + os.path.sep + isexe_nopath

# Names to test the "executable" functionality with through the
# SolverFactory.  These tests are necessary due to logic that is in
# SolverFactory that depends on whether or not a solver plugin is
# registered.
_test_names = ["cplex", "ipopt", "bonmin", "cbc", "glpk", "gurobi", "junk___"]

is_windows = os.name == 'nt'


class TestSystemCallSolver(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        import pyomo.environ

        # make sure relative path tests will work
        cls.oldpwd = os.getcwd()
        os.chdir(thisdir)
        try:
            assert exedir != exedir_user
            assert not os.path.exists(notexe_abspath_user)
            assert not os.path.exists(isexe_abspath_user)

            assert not os.path.exists(notexe_nopath)
            assert os.path.exists(os.path.join(exedir, notexe_nopath))
            assert os.path.exists(notexe_abspath)
            assert os.path.exists(notexe_relpath)
            assert not os.path.exists(isexe_nopath)
            assert os.path.exists(os.path.join(exedir, isexe_nopath))
            assert os.path.exists(isexe_abspath)
            assert os.path.exists(isexe_relpath)

            assert os.path.isfile(os.path.join(exedir, notexe_nopath))
            assert os.path.isfile(notexe_abspath)
            assert os.path.isfile(notexe_relpath)
            assert os.path.isfile(os.path.join(exedir, isexe_nopath))
            assert os.path.isfile(isexe_abspath)
            assert os.path.isfile(isexe_relpath)

            assert not os.access(os.path.join(exedir, notexe_nopath), os.X_OK)
            assert not os.access(notexe_abspath, os.X_OK)
            assert not os.access(notexe_relpath, os.X_OK)
            assert os.access(os.path.join(exedir, isexe_nopath), os.X_OK)
            assert os.access(isexe_abspath, os.X_OK)
            assert os.access(isexe_relpath, os.X_OK)
        except:
            # I don't know if tearDownClass gets called
            # if an exception occurs in setUpClass, so reset
            # the cwd here
            os.chdir(cls.oldpwd)

    @classmethod
    def tearDownClass(cls):
        os.chdir(cls.oldpwd)

    def setUp(self):
        os.chdir(thisdir)

    def test_noexe(self):
        with SystemCallSolver(type='test') as opt:
            self.assertEqual(id(opt._user_executable), id(None))
            with self.assertRaises(NotImplementedError):
                opt.executable()

    def test_available(self):
        with SystemCallSolver(type='test') as opt:
            self.assertEqual(id(opt._user_executable), id(None))
            # both cases should fail because this class is only
            # a partial implementation
            with self.assertRaises(ApplicationError):
                opt.available(exception_flag=True)
            # with self.assertRaises(ApplicationError):
            #    opt.available(exception_flag=False)

    def test_reset_executable(self):
        with SystemCallSolver(type='test') as opt:
            self.assertEqual(id(opt._user_executable), id(None))
            with self.assertRaises(NotImplementedError):
                opt.set_executable()
            opt._user_executable = 'x'
            opt.set_executable(validate=False)
            self.assertEqual(id(opt._user_executable), id(None))
            with self.assertRaises(NotImplementedError):
                opt.executable()

    def test_set_executable_notexe_nopath(self):
        with SystemCallSolver(type='test') as opt:
            self.assertEqual(id(opt._user_executable), id(None))
            with self.assertRaises(ValueError):
                opt.set_executable(notexe_nopath)
            self.assertEqual(id(opt._user_executable), id(None))
            opt.set_executable(notexe_nopath, validate=False)
            self.assertEqual(opt._user_executable, notexe_nopath)
            self.assertEqual(opt.executable(), notexe_nopath)

    @unittest.skipIf(
        is_windows,
        "Skipping test because it requires testing if a file is executable on Windows",
    )
    def test_set_executable_notexe_relpath(self):
        with SystemCallSolver(type='test') as opt:
            self.assertEqual(id(opt._user_executable), id(None))
            with self.assertRaises(ValueError):
                opt.set_executable(notexe_relpath)
            self.assertEqual(id(opt._user_executable), id(None))
            opt.set_executable(notexe_relpath, validate=False)
            self.assertEqual(opt._user_executable, notexe_relpath)
            self.assertEqual(opt.executable(), notexe_relpath)

    @unittest.skipIf(
        is_windows,
        "Skipping test because it requires testing if a file is executable on Windows",
    )
    def test_set_executable_notexe_abspath(self):
        with SystemCallSolver(type='test') as opt:
            self.assertEqual(id(opt._user_executable), id(None))
            with self.assertRaises(ValueError):
                opt.set_executable(notexe_abspath)
            self.assertEqual(id(opt._user_executable), id(None))
            opt.set_executable(notexe_abspath, validate=False)
            self.assertEqual(opt._user_executable, notexe_abspath)
            self.assertEqual(opt.executable(), notexe_abspath)

    @unittest.skipIf(
        is_windows,
        "Skipping test because it requires testing if a file is executable on Windows",
    )
    def test_set_executable_notexe_abspath_user(self):
        with SystemCallSolver(type='test') as opt:
            self.assertEqual(id(opt._user_executable), id(None))
            with self.assertRaises(ValueError):
                opt.set_executable(notexe_abspath_user)
            self.assertEqual(id(opt._user_executable), id(None))
            opt.set_executable(notexe_abspath_user, validate=False)
            self.assertEqual(opt._user_executable, notexe_abspath_user)
            self.assertEqual(opt.executable(), notexe_abspath_user)

    def test_set_executable_isexe_nopath(self):
        with SystemCallSolver(type='test') as opt:
            self.assertEqual(id(opt._user_executable), id(None))
            # because it is not in the PATH
            with self.assertRaises(ValueError):
                opt.set_executable(isexe_nopath)
            self.assertEqual(id(opt._user_executable), id(None))
            opt.set_executable(isexe_nopath, validate=False)
            self.assertEqual(opt._user_executable, isexe_nopath)
            self.assertEqual(opt.executable(), isexe_nopath)
            opt._user_executable = None

            # now add the exedir to the PATH
            rm_PATH = False
            orig_PATH = None
            if "PATH" in os.environ:
                orig_PATH = os.environ["PATH"]
            else:
                rm_PATH = True
                os.environ["PATH"] = ""
            os.environ["PATH"] = exedir + os.pathsep + os.environ["PATH"]
            try:
                opt.set_executable(isexe_nopath)
                self.assertEqual(opt._user_executable, isexe_abspath)
                self.assertEqual(opt.executable(), isexe_abspath)
            finally:
                if rm_PATH:
                    del os.environ["PATH"]
                else:
                    os.environ["PATH"] = orig_PATH

    def test_set_executable_isexe_relpath(self):
        with SystemCallSolver(type='test') as opt:
            self.assertEqual(id(opt._user_executable), id(None))
            opt.set_executable(isexe_relpath)
            self.assertEqual(opt._user_executable, isexe_abspath)
            self.assertEqual(opt.executable(), isexe_abspath)
            opt._user_executable = None
            opt.set_executable(isexe_relpath, validate=False)
            self.assertEqual(opt._user_executable, isexe_relpath)
            self.assertEqual(opt.executable(), isexe_relpath)

    def test_set_executable_isexe_abspath(self):
        with SystemCallSolver(type='test') as opt:
            self.assertEqual(id(opt._user_executable), id(None))
            opt.set_executable(isexe_abspath)
            self.assertEqual(opt._user_executable, isexe_abspath)
            self.assertEqual(opt.executable(), isexe_abspath)
            opt._user_executable = None
            opt.set_executable(isexe_abspath, validate=False)
            self.assertEqual(opt._user_executable, isexe_abspath)
            self.assertEqual(opt.executable(), isexe_abspath)

    def test_set_executable_isexe_abspath_user(self):
        with SystemCallSolver(type='test') as opt:
            self.assertEqual(id(opt._user_executable), id(None))
            opt.set_executable(isexe_abspath_user)
            self.assertTrue(os.path.samefile(opt._user_executable, isexe_abspath))
            self.assertTrue(os.path.samefile(opt.executable(), isexe_abspath))
            opt._user_executable = None
            opt.set_executable(isexe_abspath_user, validate=False)
            self.assertEqual(opt._user_executable, isexe_abspath_user)
            self.assertEqual(opt.executable(), isexe_abspath_user)

    def test_SolverFactory_executable_isexe_nopath(self):
        for name in _test_names:
            with SolverFactory(name, executable=isexe_nopath) as opt:
                self.assertTrue(isinstance(opt, UnknownSolver))

        # now add the exedir to the PATH
        rm_PATH = False
        orig_PATH = None
        if "PATH" in os.environ:
            orig_PATH = os.environ["PATH"]
        else:
            rm_PATH = True
            os.environ["PATH"] = ""
        os.environ["PATH"] = exedir + os.pathsep + os.environ["PATH"]
        try:
            for name in _test_names:
                with SolverFactory(name, executable=isexe_nopath) as opt:
                    if isinstance(opt, UnknownSolver):
                        continue
                    self.assertEqual(opt._user_executable, isexe_abspath)
                    self.assertEqual(opt.executable(), isexe_abspath)
        finally:
            if rm_PATH:
                del os.environ["PATH"]
            else:
                os.environ["PATH"] = orig_PATH

    def test_SolverFactory_executable_isexe_relpath(self):
        for name in _test_names:
            with SolverFactory(name, executable=isexe_relpath) as opt:
                if isinstance(opt, UnknownSolver):
                    continue
                self.assertEqual(opt._user_executable, isexe_abspath)
                self.assertEqual(opt.executable(), isexe_abspath)

    def test_executable_isexe_abspath(self):
        for name in _test_names:
            with SolverFactory(name, executable=isexe_abspath) as opt:
                if isinstance(opt, UnknownSolver):
                    continue
                self.assertEqual(opt._user_executable, isexe_abspath)
                self.assertEqual(opt.executable(), isexe_abspath)

    def test_executable_isexe_abspath_user(self):
        for name in _test_names:
            with SolverFactory(name, executable=isexe_abspath_user) as opt:
                if isinstance(opt, UnknownSolver):
                    continue
                # Note the user path could be different from our
                # computed path due to symlinks.  Check that the
                # executable is the same file.
                self.assertTrue(os.path.samefile(opt._user_executable, isexe_abspath))
                self.assertTrue(os.path.samefile(opt.executable(), isexe_abspath))


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# This module was originally developed as part of the PyUtilib project
# Copyright (c) 2008 Sandia Corporation.
# This software is distributed under the BSD License.
# Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
# the U.S. Government retains certain rights in this software.
# ____________________________________________________________________________________

"""Utility classes for working with loggers"""

import inspect
import io
import logging
import os
import re
import sys
import textwrap

from pyomo.version.info import releaselevel
from pyomo.common.deprecation import deprecated
from pyomo.common.fileutils import PYOMO_ROOT_DIR
from pyomo.common.flags import in_testing_environment, building_documentation
from pyomo.common.formatting import wrap_reStructuredText

_indentation_re = re.compile(r'\s*')

_RTD_URL = "https://pyomo.readthedocs.io/en/%s/errors.html"


def RTD(_id):
    _id = str(_id).lower()
    _release = (
        'stable' if releaselevel == 'final' or in_testing_environment() else 'latest'
    )
    assert _id[0] in 'wex'
    return (_RTD_URL % (_release,)) + f"#{_id}"


_DEBUG = logging.DEBUG
_NOTSET = logging.NOTSET
if not __debug__:

    def is_debug_set(logger):
        return False

elif hasattr(getattr(logging.getLogger(), 'manager', None), 'disable'):
    # This works for CPython and PyPy, but relies on a manager attribute
    # to get the current value of the logging.disabled() flag
    # (technically not included in the official logging documentation)
    def is_debug_set(logger):
        """A variant of Logger.isEnableFor that returns False if NOTSET

        The implementation of logging.Logger.isEnableFor() returns True
        if the effective level of the logger is NOTSET.  This variant
        only returns True if the effective level of the logger is NOTSET
        < level <= DEBUG.  This is used in Pyomo to detect if the user
        explicitly requested DEBUG output.

        This implementation mimics the core functionality of
        isEnabledFor() by directly querying the (undocumented) 'manager'
        attribute to get the current value for logging.disabled()

        """
        if logger.manager.disable >= _DEBUG:
            return False
        # Filter out NOTSET and higher levels
        return _NOTSET < logger.getEffectiveLevel() <= _DEBUG

elif sys.version_info[:3] < (3, 13, 4):
    # This is inefficient (it indirectly checks effective level twice),
    # but is included for [as yet unknown] platforms that ONLY implement
    # the API documented in the logging library
    def is_debug_set(logger):
        if not logger.isEnabledFor(_DEBUG):
            return False
        return logger.getEffectiveLevel() > _NOTSET

else:
    # Python 3.14 (and backported to python 3.13.4) changed the behavior
    # of isEnabledFor() so that it always returns False when called
    # while a log record is in flight (learned this from
    # https://github.com/hynek/structlog/pull/723).  In newer versions
    # of Python, we will only rely on getEffectiveLevel().
    def is_debug_set(logger):
        return _NOTSET < logger.getEffectiveLevel() <= _DEBUG


class WrappingFormatter(logging.Formatter):
    _flag = "<<!MSG!>>"

    def __init__(self, **kwds):
        if 'fmt' not in kwds:
            if kwds.get('style', '%') == '%':
                kwds['fmt'] = '%(levelname)s: %(message)s'
            elif kwds['style'] == '{':
                kwds['fmt'] = '{levelname}: {message}'
            elif kwds['style'] == '$':
                kwds['fmt'] = '$levelname: $message'
            else:
                raise ValueError('unrecognized style flag "%s"' % (kwds['style'],))
        self._wrapper = textwrap.TextWrapper(width=kwds.pop('wrap', 78))
        self._wrapper.subsequent_indent = kwds.pop('hang', ' ' * 4)
        if not self._wrapper.subsequent_indent:
            self._wrapper.subsequent_indent = ''
        self.basepath = kwds.pop('base', None)
        super(WrappingFormatter, self).__init__(**kwds)

    def format(self, record):
        msg = record.getMessage()
        if record.msg.__class__ is not str and isinstance(record.msg, Preformatted):
            return msg

        _orig = {
            k: getattr(record, k) for k in ('msg', 'args', 'pathname', 'levelname')
        }
        _id = getattr(record, 'id', None)
        record.msg = self._flag
        record.args = None
        if _id:
            record.levelname += f" ({_id.upper()})"
        if self.basepath and record.pathname.startswith(self.basepath):
            record.pathname = '[base]' + record.pathname[len(self.basepath) :]
        try:
            raw_msg = super(WrappingFormatter, self).format(record)
        finally:
            for k, v in _orig.items():
                setattr(record, k, v)

        # We want to normalize the incoming message *before* we start
        # formatting (wrapping) paragraphs.
        #
        # Most of the messages are either unformatted long lines or
        # triple-quote blocks of text.  In the latter case, if the text
        # starts on the same line as the triple-quote, then it is almost
        # certainly NOT indented with the bulk of the text, which will
        # cause dedent to get confused and not strip any leading
        # whitespace.
        #
        # A standard approach is to use inspect.cleandoc, which
        # allows for the first line to have 0 indent.
        if getattr(record, 'cleandoc', True):
            msg = inspect.cleandoc(msg)

        # Split the formatted log message (that currently has _flag in
        # lieu of the actual message content) into lines, then
        # recombine, substituting and wrapping any lines that contain
        # _flag.
        return '\n'.join(
            self._wrap_msg(line, msg, _id) if self._flag in line else line
            for line in raw_msg.splitlines()
        )

    def _wrap_msg(self, format_line, msg, _id):
        _init = self._wrapper.initial_indent, self._wrapper.subsequent_indent
        # We will honor the "hang" argument (for specifying a hanging
        # indent) unless the formatting line was indented (e.g. because
        # DEBUG was set), in which case we will use that for both the
        # first line and all subsequent lines.
        indent = _indentation_re.match(format_line).group()
        if indent:
            self._wrapper.initial_indent = self._wrapper.subsequent_indent = indent
        try:
            wrapped_msg = wrap_reStructuredText(
                format_line.strip().replace(self._flag, msg), self._wrapper
            )
        finally:
            # Restore the wrapper state
            self._wrapper.initial_indent, self._wrapper.subsequent_indent = _init
        if _id:
            wrapped_msg += f"\n{indent}{_init[1]}See also {RTD(_id)}"
        return wrapped_msg


class LegacyPyomoFormatter(logging.Formatter):
    """This mocks up the legacy Pyomo log formatting.

    This formatter takes a callback function (`verbosity`) that will be
    called for each message.  Based on the result, one of two formatting
    templates will be used.

    """

    def __init__(self, **kwds):
        if 'fmt' in kwds:
            raise ValueError("'fmt' is not a valid option for the LegacyFormatter")
        if 'style' in kwds:
            raise ValueError("'style' is not a valid option for the LegacyFormatter")

        self.verbosity = kwds.pop('verbosity', lambda: True)
        self.standard_formatter = WrappingFormatter(**kwds)
        self.verbose_formatter = WrappingFormatter(
            fmt='%(levelname)s: "%(pathname)s", %(lineno)d, %(funcName)s\n'
            '    %(message)s',
            hang=False,
            **kwds,
        )
        super(LegacyPyomoFormatter, self).__init__()

    def format(self, record):
        if self.verbosity():
            return self.verbose_formatter.format(record)
        else:
            return self.standard_formatter.format(record)


class StdoutHandler(logging.StreamHandler):
    """A logging handler that emits to the current value of sys.stdout"""

    def __init__(self):
        super().__init__()
        self.stream = None

    def flush(self):
        try:
            orig = self.stream
            self.stream = sys.stdout
            super(StdoutHandler, self).flush()
        finally:
            self.stream = orig

    def emit(self, record):
        try:
            orig = self.stream
            self.stream = sys.stdout
            super(StdoutHandler, self).emit(record)
        finally:
            self.stream = orig


class Preformatted:
    __slots__ = ('msg',)

    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return str(self.msg)

    def __repr__(self):
        return f'Preformatted({self.msg!r})'


class _GlobalLogFilter:
    def __init__(self):
        self.logger = logging.getLogger()

    def filter(self, record):
        # We will not emit messages using the default Pyomo log handler
        # if someone has registered a global handler.  However, we will
        # ignore this if we are building documentation
        # (sphinx.ext.doctest adds a handler, but we want to ignore that
        # handler when we are testing our documentation!)
        return not self.logger.handlers or building_documentation()


# This mocks up the historical Pyomo logging system, which uses a
# different formatter based on if the main pyomo logger is enabled for
# debugging.  It has been updated to suppress output if any handlers
# have been defined at the root level.
pyomo_logger = logging.getLogger('pyomo')
pyomo_handler = logging.StreamHandler(sys.stdout)
pyomo_formatter = LegacyPyomoFormatter(
    base=PYOMO_ROOT_DIR, verbosity=lambda: is_debug_set(pyomo_logger)
)
pyomo_handler.setFormatter(pyomo_formatter)
pyomo_handler.addFilter(_GlobalLogFilter())
pyomo_logger.addHandler(pyomo_handler)


@deprecated(
    'The pyomo.common.log.LogHandler class has been deprecated '
    'in favor of standard Handlers from the Python logging module '
    'combined with the pyomo.common.log.WrappingFormatter.',
    version='5.7.3',
)
class LogHandler(logging.StreamHandler):
    def __init__(self, base='', stream=None, level=logging.NOTSET, verbosity=None):
        super(LogHandler, self).__init__(stream)
        self.setLevel(level),
        if verbosity is None:
            verbosity = lambda: True
        self.setFormatter(LegacyPyomoFormatter(base=base, verbosity=verbosity))


class LoggingIntercept:
    r"""Context manager for intercepting messages sent to a log stream

    This class is designed to enable easy testing of log messages.

    The LoggingIntercept context manager will intercept messages sent to
    a log stream matching a specified level and send the messages to the
    specified output stream.  Other handlers registered to the target
    logger will be temporarily removed and the logger will be set not to
    propagate messages up to higher-level loggers.

    Parameters
    ----------
    output: io.TextIOBase
        the file stream to send log messages to

    module: str
        the target logger name to intercept. `logger` and `module` are
        mutually exclusive.

    level: int
        the logging level to intercept

    formatter: logging.Formatter
        the formatter to use when rendering the log messages.  If not
        specified, uses `'%(message)s'`

    logger: logging.Logger
        the target logger to intercept. `logger` and `module` are
        mutually exclusive.

    Examples
    --------
    >>> import io, logging
    >>> from pyomo.common.log import LoggingIntercept
    >>> buf = io.StringIO()
    >>> with LoggingIntercept(buf, 'pyomo.core', logging.WARNING):
    ...     logging.getLogger('pyomo.core').warning('a simple message')
    >>> buf.getvalue()
    'a simple message\n'

    """

    def __init__(
        self,
        output=None,
        module=None,
        level=logging.WARNING,
        formatter=None,
        logger=None,
    ):
        self.handler = None
        self.output = output
        if logger is not None:
            if module is not None:
                raise ValueError(
                    "LoggingIntercept: only one of 'module' and 'logger' is allowed"
                )
            self._logger = logger
        else:
            self._logger = logging.getLogger(module)
        self._level = level
        if formatter is None:
            formatter = logging.Formatter('%(message)s')
        self._formatter = formatter
        self._save = None

    def __enter__(self):
        # Get the logger for the scope we will be overriding
        logger = self._logger
        self._save = logger.level, logger.propagate, logger.handlers
        if self._level is None:
            self._level = logger.getEffectiveLevel()
        # Set up the handler
        output = self.output
        if output is None:
            output = io.StringIO()
        assert self.handler is None
        self.handler = logging.StreamHandler(output)
        self.handler.setFormatter(self._formatter)
        self.handler.setLevel(self._level)
        # Register the handler with the appropriate module scope
        logger.handlers = []
        logger.propagate = False
        logger.setLevel(self.handler.level)
        logger.addHandler(self.handler)
        return output

    def __exit__(self, et, ev, tb):
        logger = self._logger
        logger.removeHandler(self.handler)
        self.handler = None
        logger.setLevel(self._save[0])
        logger.propagate = self._save[1]
        assert not logger.handlers
        logger.handlers.extend(self._save[2])

    @property
    def module(self):
        return self._logger.name


class LogStream(io.TextIOBase):
    """
    This class logs whatever gets sent to the write method.
    This is useful for logging solver output (a LogStream
    instance can be handed to TeeStream from pyomo.common.tee).
    """

    def __init__(self, level, logger):
        self._level = level
        self._logger = logger
        self._buffer = ''

    def write(self, s: str) -> int:
        if not s:
            return 0
        res = len(s)
        if self._buffer:
            s = self._buffer + s
        lines = s.split('\n')
        for line in lines[:-1]:
            self._logger.log(self._level, line)
        self._buffer = lines[-1]
        return res

    def flush(self):
        if self._buffer:
            self.write('\n')

    def redirect_streams(self, redirects):
        """Redirect StreamHandler objects to the original file descriptors

        This utility method for py:class:`~pyomo.common.tee.capture_output`
        locates any StreamHandlers that would process messages from the
        logger assigned to this :py:class:`LogStream` that would write
        to the file descriptors redirected by `capture_output` and
        yields context managers that will redirect those StreamHandlers
        back to duplicates of the original file descriptors.

        """
        found = 0
        logger = self._logger
        while isinstance(logger, logging.LoggerAdapter):
            logger = logger.logger
        while logger:
            for handler in logger.handlers:
                found += 1
                if not isinstance(handler, logging.StreamHandler):
                    continue
                try:
                    fd = handler.stream.fileno()
                except (AttributeError, OSError):
                    fd = None
                if fd not in redirects:
                    continue
                yield _StreamRedirector(handler, redirects[fd].original_fd)
            if not logger.propagate:
                break
            else:
                logger = logger.parent
        if not found:
            fd = logging.lastResort.stream.fileno()
            if not redirects:
                yield _LastResortRedirector(fd)
            elif fd in redirects:
                yield _LastResortRedirector(redirects[fd].original_fd)


class _StreamRedirector:
    def __init__(self, handler, fd):
        self.handler = handler
        self.fd = fd
        self.local_fd = None
        self.orig_stream = None

    def __enter__(self):
        assert self.local_fd is None
        self.orig_stream = self.handler.stream
        # Note: ideally, we would use closefd=True and let Python handle
        # closing the local file descriptor that we are about to create.
        # However, it appears that closefd is ignored on Windows (see
        # #3587), so we will just handle it explicitly ourselves.
        self.local_fd = os.dup(self.fd)
        self.handler.stream = os.fdopen(
            self.local_fd, mode="w", closefd=False
        ).__enter__()

    def __exit__(self, et, ev, tb):
        try:
            self.handler.stream.__exit__(et, ev, tb)
            os.close(self.local_fd)
        finally:
            self.handler.stream = self.orig_stream


class _LastResortRedirector:
    def __init__(self, fd):
        self.fd = fd
        self.local_fd = None
        self.orig_stream = None

    def __enter__(self):
        assert self.local_fd is None
        self.orig = logging.lastResort
        # Note: ideally, we would use closefd=True and let Python handle
        # closing the local file descriptor that we are about to create.
        # However, it appears that closefd is ignored on Windows (see
        # #3587), so we will just handle it explicitly ourselves.
        self.local_fd = os.dup(self.fd)
        logging.lastResort = logging.StreamHandler(
            os.fdopen(self.local_fd, mode="w", closefd=False).__enter__()
        )

    def __exit__(self, et, ev, tb):
        try:
            logging.lastResort.stream.close()
            os.close(self.local_fd)
        finally:
            logging.lastResort = self.orig



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.core import TransformationFactory, Var, NonNegativeReals
from pyomo.core.base.misc import create_name

from pyomo.core.plugins.transform.hierarchy import IsomorphicTransformation
from pyomo.core.plugins.transform.util import collectAbstractComponents


@TransformationFactory.register(
    "core.add_slack_vars",
    doc="Create an equivalent model by introducing slack variables to eliminate inequality constraints.",
)
class EqualityTransform(IsomorphicTransformation):
    """
    Creates a new, equivalent model by introducing slack and excess variables
    to eliminate inequality constraints.
    """

    def __init__(self, **kwds):
        kwds["name"] = kwds.pop("name", "add_slack_vars")
        super(EqualityTransform, self).__init__(**kwds)

    def _create_using(self, model, **kwds):
        """
        Eliminate inequality constraints.

        Required arguments:

          model The model to transform.

        Optional keyword arguments:

          slack_root  The root name of auxiliary slack variables.
                      Default is 'auxiliary_slack'.
          excess_root The root name of auxiliary slack variables.
                      Default is 'auxiliary_excess'.
          lb_suffix   The suffix applied to converted upper bound constraints
                      Default is '_lower_bound'.
          ub_suffix   The suffix applied to converted lower bound constraints
                      Default is '_upper_bound'.
        """

        # Optional naming schemes
        slack_suffix = kwds.pop("slack_suffix", "slack")
        excess_suffix = kwds.pop("excess_suffix", "excess")
        lb_suffix = kwds.pop("lb_suffix", "lb")
        ub_suffix = kwds.pop("ub_suffix", "ub")

        equality = model.clone()
        components = collectAbstractComponents(equality)

        #
        # Fix all Constraint objects
        #
        for con_name in components["Constraint"]:
            con = equality.__getattribute__(con_name)

            #
            # Get all ConstraintData objects
            #
            # We need to get the keys ahead of time because we are modifying
            # con._data on-the-fly.
            #
            indices = con._data.keys()
            for ndx, cdata in [(ndx, con._data[ndx]) for ndx in indices]:
                qualified_con_name = create_name(con_name, ndx)

                # Do nothing with equality constraints
                if cdata.equality:
                    continue

                # Add an excess variable if the lower bound exists
                if cdata.lower is not None:
                    # Make the excess variable
                    excess_name = "%s_%s" % (qualified_con_name, excess_suffix)
                    equality.__setattr__(excess_name, Var(within=NonNegativeReals))

                    # Make a new lower bound constraint
                    lb_name = "%s_%s" % (create_name("", ndx), lb_suffix)
                    excess = equality.__getattribute__(excess_name)
                    new_expr = cdata.lower == cdata.body - excess
                    con.add(lb_name, new_expr)

                # Add a slack variable if the lower bound exists
                if cdata.upper is not None:
                    # Make the excess variable
                    slack_name = "%s_%s" % (qualified_con_name, slack_suffix)
                    equality.__setattr__(slack_name, Var(within=NonNegativeReals))

                    # Make a new upper bound constraint
                    ub_name = "%s_%s" % (create_name("", ndx), ub_suffix)
                    slack = equality.__getattribute__(slack_name)
                    new_expr = cdata.upper == cdata.body + slack
                    con.add(ub_name, new_expr)

                # Since we explicitly `continue` for equality constraints, we
                # can safely remove the old ConstraintData object
                del con._data[ndx]

        return equality.create()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""Functions for initializing the main problem
in Logic-based outer approximation.
"""

from contextlib import contextmanager
from math import fabs

from pyomo.common.collections import ComponentMap

from pyomo.contrib.gdpopt.cut_generation import add_no_good_cut
from pyomo.contrib.gdpopt.solve_discrete_problem import solve_MILP_discrete_problem
from pyomo.contrib.gdpopt.util import _DoNothing
from pyomo.core import Block, Constraint, Objective, Var, maximize, value
from pyomo.gdp import Disjunct
from pyomo.opt import TerminationCondition as tc


def _collect_original_bounds(discrete_prob_util_block):
    original_bounds = ComponentMap()
    for v in discrete_prob_util_block.all_mip_variables:
        original_bounds[v] = (v.lb, v.ub)
    return original_bounds


def _restore_bounds(original_bounds):
    for v, (l, u) in original_bounds.items():
        v.setlb(l)
        v.setub(u)


# This contextmanager is for use when we solve the discrete problem with some
# variables fixed. In that case, the bounds tightening that might be done during
# preprocessing is not valid later, and we need to restore the variable bounds.
@contextmanager
def preserve_discrete_problem_feasible_region(
    discrete_problem_util_block, config, original_bounds=None
):
    if config.mip_presolve and original_bounds is None:
        original_bounds = _collect_original_bounds(discrete_problem_util_block)

    yield

    if config.mip_presolve:
        _restore_bounds(original_bounds)


def init_custom_disjuncts(
    util_block, discrete_problem_util_block, subprob_util_block, config, solver
):
    """Initialize by using user-specified custom disjuncts."""
    solver._log_header(config.logger)

    used_disjuncts = {}

    # We are going to fix indicator_vars in the discrete problem before we
    # solve it, so the bounds tightening will not necessarily be valid
    # afterward. So we save these bounds and restore them in each iteration. We
    # collect them here since there's no point in doing that every iteration.
    if config.mip_presolve:
        original_bounds = _collect_original_bounds(discrete_problem_util_block)

    for count, active_disjunct_set in enumerate(config.custom_init_disjuncts):
        # custom_init_disjuncts contains a list of sets, giving the disjuncts
        # active at each initialization iteration
        used_disjuncts = set()

        subproblem = subprob_util_block.parent_block()
        # fix the disjuncts in the linear GDP and solve
        config.logger.info(
            "Generating initial linear GDP approximation by "
            "solving subproblems with user-specified active disjuncts."
        )
        for orig_disj, discrete_problem_disj in zip(
            util_block.disjunct_list, discrete_problem_util_block.disjunct_list
        ):
            if orig_disj in active_disjunct_set:
                used_disjuncts.add(orig_disj)
                discrete_problem_disj.indicator_var.fix(True)
            else:
                discrete_problem_disj.indicator_var.fix(False)
        unused = set(active_disjunct_set) - used_disjuncts
        if len(unused) > 0:
            config.logger.warning(
                'The following disjuncts from the custom disjunct '
                'initialization set number %s were unused: '
                '%s\nThey may not be Disjunct objects or '
                'they may not be on the active subtree being '
                'solved.' % (count, ", ".join([disj.name for disj in unused]))
            )
        with preserve_discrete_problem_feasible_region(
            discrete_problem_util_block, config, original_bounds
        ):
            mip_termination = solve_MILP_discrete_problem(
                discrete_problem_util_block, solver, config
            )
        if mip_termination is not tc.infeasible:
            solver._fix_discrete_soln_solve_subproblem_and_add_cuts(
                discrete_problem_util_block, subprob_util_block, config
            )
            # remove the integer solution
            add_no_good_cut(discrete_problem_util_block, config)
        else:
            config.logger.error(
                'MILP relaxation infeasible for user-specified '
                'custom initialization disjunct set %s. '
                'Skipping that set and continuing on.'
                % list(disj.name for disj in active_disjunct_set)
            )
        solver.initialization_iteration += 1


def init_fixed_disjuncts(
    util_block, discrete_problem_util_block, subprob_util_block, config, solver
):
    """Initialize by solving the problem with the current disjunct values."""

    config.logger.info(
        "Generating initial linear GDP approximation by "
        "solving subproblem with original user-specified disjunct values."
    )
    solver._log_header(config.logger)

    # Again, if we presolve, we are going to tighten the bounds after fixing the
    # indicator_vars, so it won't be valid afterwards and we need to restore it.
    with preserve_discrete_problem_feasible_region(discrete_problem_util_block, config):
        # fix the disjuncts in the discrete problem and send for solution.
        already_fixed = set()
        for disj in discrete_problem_util_block.disjunct_list:
            indicator = disj.indicator_var
            if indicator.fixed:
                already_fixed.add(disj)
            else:
                indicator.fix()

        # We copied the variables over when we cloned, and because the Booleans
        # are auto-linked to the binaries, we shouldn't have to change
        # anything. So first we solve the discrete problem in case we need
        # values for other discrete variables, and to make sure it's feasible.
        mip_termination = solve_MILP_discrete_problem(
            discrete_problem_util_block, solver, config
        )

        # restore the fixed status of the indicator_variables
        for disj in discrete_problem_util_block.disjunct_list:
            if disj not in already_fixed:
                disj.indicator_var.unfix()

    if mip_termination is not tc.infeasible:
        solver._fix_discrete_soln_solve_subproblem_and_add_cuts(
            discrete_problem_util_block, subprob_util_block, config
        )
        add_no_good_cut(discrete_problem_util_block, config)
    else:
        config.logger.error(
            'MILP relaxation infeasible for initial user-specified '
            'disjunct values. '
            'Skipping initialization.'
        )
    solver.initialization_iteration += 1


@contextmanager
def use_discrete_problem_for_max_binary_initialization(discrete_problem_util_block):
    m = discrete_problem_util_block.parent_block()

    # Set up binary maximization objective
    original_objective = next(
        m.component_data_objects(Objective, active=True, descend_into=True)
    )
    original_objective.deactivate()

    binary_vars = (
        v
        for v in m.component_data_objects(ctype=Var, descend_into=(Block, Disjunct))
        if v.is_binary() and not v.fixed
    )
    discrete_problem_util_block.max_binary_obj = Objective(
        expr=sum(binary_vars), sense=maximize
    )

    yield

    # clean up the objective. We don't clean up the no-good cuts because we
    # still want them. We've already considered those solutions.
    del discrete_problem_util_block.max_binary_obj
    original_objective.activate()


def init_max_binaries(
    util_block, discrete_problem_util_block, subprob_util_block, config, solver
):
    """Initialize by maximizing binary variables and disjuncts.

    This function activates as many binary variables and disjuncts as
    feasible.

    """
    config.logger.info(
        "Generating initial linear GDP approximation by "
        "solving a subproblem that maximizes "
        "the sum of all binary and logical variables."
    )
    solver._log_header(config.logger)

    # As with set covering, this is only a change of objective. The formulation
    # may be tightened, but that is valid for the duration.
    with use_discrete_problem_for_max_binary_initialization(
        discrete_problem_util_block
    ):
        mip_termination = solve_MILP_discrete_problem(
            discrete_problem_util_block, solver, config
        )
        if mip_termination is not tc.infeasible:
            solver._fix_discrete_soln_solve_subproblem_and_add_cuts(
                discrete_problem_util_block, subprob_util_block, config
            )
        else:
            config.logger.debug(
                "MILP relaxation for initialization was infeasible. "
                "Problem is infeasible."
            )
            solver._update_dual_bound_to_infeasible()
            return False
        add_no_good_cut(discrete_problem_util_block, config)

    solver.initialization_iteration += 1


@contextmanager
def use_discrete_problem_for_set_covering(discrete_problem_util_block):
    m = discrete_problem_util_block.parent_block()

    original_objective = next(
        m.component_data_objects(Objective, active=True, descend_into=True)
    )
    original_objective.deactivate()
    # placeholder for the objective
    discrete_problem_util_block.set_cover_obj = Objective(expr=0, sense=maximize)

    yield

    # clean up the objective. We don't clean up the no-good cuts because we
    # still want them. We've already considered those solutions.
    del discrete_problem_util_block.set_cover_obj
    original_objective.activate()


def update_set_covering_objective(discrete_problem_util_block, disj_needs_cover):
    # number of disjuncts that still need to be covered
    num_needs_cover = sum(1 for disj_bool in disj_needs_cover if disj_bool)
    # number of disjuncts that have been covered
    num_covered = len(disj_needs_cover) - num_needs_cover
    # weights for the set covering problem
    weights = list(
        (num_covered + 1 if disj_bool else 1) for disj_bool in disj_needs_cover
    )
    # Update set covering objective
    discrete_problem_util_block.set_cover_obj.expr = sum(
        weight * disj.binary_indicator_var
        for (weight, disj) in zip(weights, discrete_problem_util_block.disjunct_list)
    )


def init_set_covering(
    util_block, discrete_problem_util_block, subprob_util_block, config, solver
):
    """Initialize by solving problems to cover the set of all disjuncts.

    The purpose of this initialization is to generate linearizations
    corresponding to each of the disjuncts.

    This work is based upon prototyping work done by Eloy Fernandez at
    Carnegie Mellon University.

    """
    config.logger.info("Starting set covering initialization.")
    solver._log_header(config.logger)

    # List of True/False if the corresponding disjunct in
    # disjunct_list still needs to be covered by the initialization
    disjunct_needs_cover = list(
        any(
            constr.body.polynomial_degree() not in (0, 1)
            for constr in disj.component_data_objects(
                ctype=Constraint, active=True, descend_into=True
            )
        )
        for disj in util_block.disjunct_list
    )
    subprob = subprob_util_block.parent_block()

    # We borrow the discrete problem to be the set covering MIP. This is only a
    # change of objective. The formulation may have its bounds tightened as a
    # result of preprocessing in the MIP solves, but that is okay because the
    # feasible region is the same as the original discrete problem and any
    # feasibility-based tightening will remain valid for the duration.
    with use_discrete_problem_for_set_covering(discrete_problem_util_block):
        iter_count = 1
        while any(disjunct_needs_cover) and iter_count <= config.set_cover_iterlim:
            config.logger.debug(
                "%s disjuncts need to be covered." % disjunct_needs_cover.count(True)
            )
            ## Solve set covering MIP
            update_set_covering_objective(
                discrete_problem_util_block, disjunct_needs_cover
            )

            mip_termination = solve_MILP_discrete_problem(
                discrete_problem_util_block, solver, config
            )

            if mip_termination is tc.infeasible:
                config.logger.debug(
                    'Set covering problem is infeasible. '
                    'Problem may have no more feasible '
                    'disjunctive realizations.'
                )
                if iter_count <= 1:
                    config.logger.warning(
                        'Set covering problem is infeasible. '
                        'Check your linear and logical constraints '
                        'for contradictions.'
                    )
                solver._update_dual_bound_to_infeasible()
                # problem is infeasible. break
                return False
            else:
                config.logger.debug('Solved set covering MIP')

            ## solve local NLP
            nlp_feasible = solver._fix_discrete_soln_solve_subproblem_and_add_cuts(
                discrete_problem_util_block, subprob_util_block, config
            )
            if nlp_feasible:
                # if successful, update sets
                active_disjuncts = list(
                    fabs(value(disj.binary_indicator_var) - 1)
                    <= config.integer_tolerance
                    for disj in discrete_problem_util_block.disjunct_list
                )
                # Update the disjunct needs cover list
                disjunct_needs_cover = list(
                    (needed_cover and not was_active)
                    for (needed_cover, was_active) in zip(
                        disjunct_needs_cover, active_disjuncts
                    )
                )
            add_no_good_cut(discrete_problem_util_block, config)
            iter_count += 1
            solver.initialization_iteration += 1

        if any(disjunct_needs_cover):
            # Iteration limit was hit without a full covering of all nonlinear
            # disjuncts
            config.logger.warning(
                'Iteration limit reached for set covering initialization '
                'without covering all disjuncts.'
            )
            return False

    config.logger.info("Initialization complete.")
    return True


# Valid initialization strategies
valid_init_strategies = {
    'no_init': _DoNothing,
    'set_covering': init_set_covering,
    'max_binary': init_max_binaries,
    'fix_disjuncts': init_fixed_disjuncts,
    'custom_disjuncts': init_custom_disjuncts,
}



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Unit Tests for pyomo.base.misc
#

import glob
import importlib
import os
import subprocess
import sys

from itertools import filterfalse
from os.path import join

import pyomo.common.dependencies as dependencies
from pyomo.common.fileutils import PYOMO_ROOT_DIR

import pyomo.common.unittest as unittest

parameterized, param_available = dependencies.attempt_import('parameterized')
parameterized = parameterized.parameterized

_FAST_TEST = False

# List of directories under `pyomo` that intentionally do NOT have
# __init__.py files (because they either contain no Python files - or
# contain Python files that are only used in testing and explicitly NOT
# part of the "Pyomo package")
_NON_MODULE_DIRS = {
    join('contrib', 'ampl_function_demo', 'src'),
    join('contrib', 'appsi', 'cmodel', 'src'),
    join('contrib', 'cspline_external', 'src'),
    join('contrib', 'aslfunctions', 'src'),
    join('contrib', 'simplification', 'ginac', 'src'),
    join('contrib', 'pynumero', 'src'),
    join('core', 'tests', 'data', 'baselines'),
    join('core', 'tests', 'diet', 'baselines'),
    join('opt', 'tests', 'solver', 'exe_dir'),
    join('solvers', 'tests', 'checks', 'data'),
    join('solvers', 'tests', 'mip', 'asl'),
    join('solvers', 'tests', 'mip', 'cbc'),
    join('solvers', 'tests', 'mip', 'cplex'),
    join('solvers', 'tests', 'mip', 'glpk'),
    join('solvers', 'tests', 'piecewise_linear', 'baselines'),
    join('solvers', 'tests', 'piecewise_linear', 'kernel_baselines'),
    join('solvers', 'tests', 'piecewise_linear', 'kernel_problems'),
    join('solvers', 'tests', 'piecewise_linear', 'problems'),
}

_DO_NOT_IMPORT_MODULES = {
    'pyomo.common.tests.dep_mod_except',
    'pyomo.contrib.interior_point.examples.ex1',
}

try:
    _cwd = os.getcwd()
    os.chdir(os.path.join(PYOMO_ROOT_DIR, 'pyomo'))
    modules = sorted(
        os.path.join('pyomo', os.path.splitext(fname)[0]).replace(os.path.sep, '.')
        for fname in glob.glob(os.path.join('**', '*.py'), recursive=True)
        if not fname.endswith('__init__.py')
    )
    modules = list(filterfalse(_DO_NOT_IMPORT_MODULES.__contains__, modules))
finally:
    os.chdir(_cwd)


import_test = """
import pyomo.common.dependencies
pyomo.common.dependencies.SUPPRESS_DEPENDENCY_WARNINGS = True
import unittest
try:
    import %s
except unittest.case.SkipTest as e:
    # suppress the exception, but print the message
    print(e)
"""


class TestPackageLayout(unittest.TestCase):
    def test_for_init_files(self):
        _NMD = set(_NON_MODULE_DIRS)
        fail = []
        module_dir = os.path.join(PYOMO_ROOT_DIR, 'pyomo')
        for path, subdirs, files in os.walk(module_dir):
            assert path.startswith(module_dir)
            relpath = path[1 + len(module_dir) :]
            # Skip all __pycache__ directories
            try:
                subdirs.remove('__pycache__')
            except ValueError:
                pass

            if '__init__.py' in files:
                continue

            if relpath in _NMD:
                _NMD.remove(relpath)
                # Skip checking all subdirectories
                subdirs[:] = []
                continue

            fail.append(relpath)

        if fail:
            self.fail(
                "Directories are missing __init__.py files:\n\t"
                + "\n\t".join(sorted(fail))
            )
        if _NMD:
            self.fail(
                "_NON_MODULE_DIRS contains entries not found in package "
                "or unexpectedly contain a __init__.py file:\n\t"
                + "\n\t".join(sorted(_NMD))
            )

    @parameterized.expand(modules)
    @unittest.pytest.mark.importtest
    def test_module_import(self, module):
        # We will go through the entire package and ensure that all the
        # python modules are a least importable.  This is important to
        # be tested on the newest Python version (in part to catch
        # deprecation warnings before they become fatal parse errors).
        module_file = (
            os.path.join(PYOMO_ROOT_DIR, module.replace('.', os.path.sep)) + '.py'
        )
        # we need to delete the .pyc file, because some things (like
        # invalid docstrings) only toss the warning when the module is
        # initially byte-compiled.
        pyc = importlib.util.cache_from_source(module_file)
        if os.path.isfile(pyc):
            os.remove(pyc)
        test_code = import_test % module
        if _FAST_TEST:
            # This is much faster, as it only reloads each module once
            # (no subprocess, and no reloading and dependent modules).
            # However, it will generate false positives when reimporting
            # a single module creates side effects (this happens in some
            # of the testing harness for auto-registered test cases)
            from pyomo.common.fileutils import import_file
            import warnings

            try:
                _dep_warn = dependencies.SUPPRESS_DEPENDENCY_WARNINGS
                dependencies.SUPPRESS_DEPENDENCY_WARNINGS = True

                with warnings.catch_warnings():
                    warnings.resetwarnings()
                    warnings.filterwarnings('error')
                    import_file(module_file, clear_cache=True)
            except unittest.SkipTest as e:
                # suppress the exception, but print the message
                print(e)
            finally:
                dependencies.SUPPRESS_DEPENDENCY_WARNINGS = _dep_warn
        else:
            subprocess.run([sys.executable, '-Werror', '-c', test_code], check=True)
